# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import csv
import datetime
import logging
import os
import os.path

import dateutil.parser
import six
from celery import shared_task
from django.conf import settings
from django.db import IntegrityError, transaction
from django.db.models import Count, Sum

from karaage.institutes.models import Institute
from karaage.machines.models import Machine
from karaage.projects.models import Project

from . import graphs, usage
from .dirs import GRAPH_ROOT, GRAPH_TMP
from .models import (
    CPUJob,
    InstituteCache,
    MachineCache,
    MachineCategoryCache,
    PersonCache,
    ProjectCache,
)


import matplotlib  # isort:skip
os.environ['MPLCONFIGDIR'] = GRAPH_TMP
matplotlib.use('Agg')
import matplotlib.pyplot as plt  # NOQA isort:skip
import matplotlib.dates as mdates  # NOQA isort:skip

logger = logging.getLogger(__name__)


@shared_task
def gen_machine_category_cache(start, end):
    start = dateutil.parser.parse(start).date()
    end = dateutil.parser.parse(end).date()

    current = gen_machine_category_cache

    logger.info("gen_machine_category_cache")

    total = 3
    i = 0

    logger.info("caching machines")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching machines', })
    _gen_machine_cache(start, end)
    i = i + 1

    logger.info("caching categories")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching categories', })
    _gen_machine_category_cache(start, end)
    i = i + 1

    logger.info("generating category graphs")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'generating category graphs'})
    _gen_machine_graph(start, end, force_overwrite=False)
    i = i + 1

    logger.info("finished")


@shared_task
def gen_cache_for_machine_category(start, end):
    start = dateutil.parser.parse(start).date()
    end = dateutil.parser.parse(end).date()

    current = gen_cache_for_machine_category

    total = 6
    i = 0

    logger.info("caching machines")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching machines', })
    _gen_machine_cache(start, end)
    i = i + 1

    logger.info("caching categories")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching categories', })
    _gen_machine_category_cache(start, end)
    i = i + 1

    logger.info("caching institutes")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching institutes'})
    _gen_institute_cache(start, end)
    i = i + 1

    logger.info("caching projects")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching projects'})
    _gen_project_cache(start, end)
    i = i + 1

    logger.info("caching people")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'caching people'})
    _gen_person_cache(start, end)
    i = i + 1

    logger.info("generating machine category graph")
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'generating machine category graphs', })
    _gen_institute_graph(start, end, force_overwrite=False)
    _gen_trend_graph(start, end, force_overwrite=False)
    i = i + 1

    logger.info("finished")


@shared_task
def gen_cache_for_project(start, end, project_pk):
    start = dateutil.parser.parse(start).date()
    end = dateutil.parser.parse(end).date()

    project = Project.objects.get(pk=project_pk)

    current = gen_cache_for_project
    total = 1
    i = 0

    logger.info("generating project graphs %s" % project)
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'generating project graphs'})
    _gen_project_trend_graph(project, start, end, force_overwrite=False)
    i = i + 1


@shared_task
def gen_cache_for_institute(start, end, institute_pk):
    start = dateutil.parser.parse(start).date()
    end = dateutil.parser.parse(end).date()

    institute = Institute.objects.get(pk=institute_pk)

    current = gen_cache_for_institute
    total = 1
    i = 0

    logger.info("generating institute graphs %s" % institute)
    current.update_state(
        state='PROGRESS',
        meta={'completed': i, 'total': total,
              'message': 'generating institute graphs'})
    _gen_institute_trend_graph(institute, start, end, force_overwrite=False)
    i = i + 1


@shared_task
def gen_cache_for_all_institutes(start, end):
    start = dateutil.parser.parse(start).date()
    end = dateutil.parser.parse(end).date()

    current = gen_cache_for_all_institutes
    total = Institute.objects.count()
    i = 0

    for institute in Institute.objects.all():
        logger.info("generating institute graphs %s" % institute)
        current.update_state(
            state='PROGRESS',
            meta={'completed': i, 'total': total,
                  'message': 'generating institute graphs'})
        _gen_institute_trend_graph(institute, start, end, force_overwrite=False)
        i = i + 1


# -----------------------------------------------------------------------


def _gen_machine_category_cache(start, end):
    try:
        with transaction.atomic():
            cache = MachineCategoryCache.objects.create(
                start=start, end=end,
                cpu_time=0, no_jobs=0,
                available_time=0)
            machines = Machine.objects.all()
            total_time = 0
            for machine in machines:
                m_start = machine.start_date
                m_end = machine.end_date
                if not m_end:
                    m_end = end

                assert start <= end
                assert m_start <= m_end

                if start > m_end or m_start > end:
                    continue
                if end < m_start or m_end < start:
                    continue

                this_start = start
                if start < m_start:
                    this_start = m_start
                this_end = end
                if end > m_end:
                    this_end = m_end

                num_days = (this_end - this_start).days + 1
                total_time += (machine.no_cpus * num_days * 24 * 60 * 60)

            query = CPUJob.objects.filter(
                date__range=(start, end))
            data = query.aggregate(
                usage=Sum('cpu_usage'),
                jobs=Count('id'))

            data['usage'] = data['usage'] or 0

            cache.cpu_time = data['usage']
            cache.no_jobs = data['jobs']
            cache.available_time = total_time
            cache.save()
    except IntegrityError:
        # entry already exists
        pass


def _gen_machine_cache(start, end):
    query = CPUJob.objects.filter(date__range=(start, end))
    query = query.values('machine')
    query = query.annotate(usage=Sum('cpu_usage'), jobs=Count('id'))
    query = query.order_by()

    for data in query.iterator():
        data['usage'] = data['usage'] or 0
        machine = data['machine']
        try:
            with transaction.atomic():
                MachineCache.objects.create(
                    machine_id=machine,
                    start=start, end=end,
                    cpu_time=data['usage'], no_jobs=data['jobs'])
        except IntegrityError:
            # entry already exists
            pass


def _gen_institute_cache(start, end):
    query = CPUJob.objects.filter(
        date__range=(start, end),
    )
    query = query.values('project__institute')
    query = query.annotate(usage=Sum('cpu_usage'), jobs=Count('id'))
    query = query.order_by()

    for data in query.iterator():
        data['usage'] = data['usage'] or 0
        institute = data['project__institute']
        try:
            with transaction.atomic():
                InstituteCache.objects.create(
                    institute_id=institute,
                    start=start, end=end,
                    cpu_time=data['usage'], no_jobs=data['jobs'])
        except IntegrityError:
            # entry already exists
            pass


def _gen_project_cache(start, end):
    query = CPUJob.objects.filter(
        date__range=(start, end),
    )
    query = query.values('project')
    query = query.annotate(usage=Sum('cpu_usage'), jobs=Count('id'))
    query = query.order_by()

    for data in query.iterator():
        data['usage'] = data['usage'] or 0
        project = data['project']
        try:
            with transaction.atomic():
                ProjectCache.objects.create(
                    project_id=project,
                    start=start, end=end,
                    cpu_time=data['usage'], no_jobs=data['jobs'])
        except IntegrityError:
            # entry already exists
            pass


def _gen_person_cache(start, end):
    query = CPUJob.objects.filter(
        date__range=(start, end),
    )
    query = query.values('project', 'account__person')
    query = query.annotate(usage=Sum('cpu_usage'), jobs=Count('id'))
    query = query.order_by()

    for data in query.iterator():
        data['usage'] = data['usage'] or 0
        project = data['project']
        person = data['account__person']
        try:
            with transaction.atomic():
                PersonCache.objects.create(
                    person_id=person, project_id=project,
                    start=start, end=end,
                    cpu_time=data['usage'], no_jobs=data['jobs'])
        except IntegrityError:
            # entry already exists
            pass


# -----------------------------------------------------------------------

def _check_directory_exists(filename):
    base_path = os.path.dirname(filename)
    if not os.path.exists(base_path):
        os.makedirs(base_path)


def _gen_project_trend_graph(project, start, end, force_overwrite=False):
    """Generates a bar graph for a project

    Keyword arguments:
    project -- Project
    start -- start date
    end -- end date

    """
    filename = graphs.get_project_trend_graph_filename(project, start, end)
    csv_filename = os.path.join(GRAPH_ROOT, filename + '.csv')
    png_filename = os.path.join(GRAPH_ROOT, filename + '.png')

    _check_directory_exists(csv_filename)
    _check_directory_exists(png_filename)

    if not settings.GRAPH_DEBUG or force_overwrite:
        if os.path.exists(csv_filename):
            if os.path.exists(png_filename):
                return

    query = CPUJob.objects.filter(
        project=project,
        date__range=(start, end)
    )
    query = query.values('account', 'account__username', 'date')
    query = query.annotate(Sum('cpu_usage')).order_by('account', 'date')

    t_start = start
    t_end = end

    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')

    fig, ax = plt.subplots(figsize=(6, 4))
    ax.set_xlim(start, end + datetime.timedelta(days=1))
    ax.set_title('%s   %s - %s' % (project.pid, start_str, end_str))
    ax.set_ylabel("CPU Time (hours)")
    ax.set_xlabel("Date")

    locator = mdates.AutoDateLocator()
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(locator))
    ax.xaxis.set_minor_locator(mdates.DayLocator())

    data = {}
    x_data = {}
    y_data = {}

    with open(csv_filename, 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        for row in query.iterator():
            csv_writer.writerow([
                row['account__username'],
                row['date'], row['cpu_usage__sum'] / 3600.00
            ])

            account = row['account']
            date = row['date']

            if account not in data:
                data[account] = {}
                x_data[account] = []
                y_data[account] = []

            data[account][date] = row['cpu_usage__sum']

    for account, dates in six.iteritems(data):
        start = t_start
        end = t_end
        while start <= end:
            total = 0
            if start in dates:
                total = dates[start]
            x_data[account].append(start)
            y_data[account].append(total / 3600.00)
            start = start + datetime.timedelta(days=1)

    del data

    totals = []
    start = t_start
    end = t_end
    while start <= end:
        totals.append(0)
        start = start + datetime.timedelta(days=1)

    count = 0
    for account in x_data.keys():
        ax.bar(
            x_data[account], y_data[account],
            bottom=totals,
            color=graphs.get_colour(count),
            edgecolor=graphs.get_colour(count),
            align='edge')
        count = count + 1

        i = 0
        start = t_start
        end = t_end
        while start <= end:
            totals[i] += y_data[account][i]
            i = i + 1
            start = start + datetime.timedelta(days=1)

    del x_data
    del y_data
    del totals

    fig.autofmt_xdate()
    plt.tight_layout()
    plt.savefig(png_filename)
    plt.close()


def _gen_institute_graph(start, end, force_overwrite=False):
    """ Pie chart comparing institutes usage. """
    filename = graphs.get_institute_graph_filename(start, end)
    csv_filename = os.path.join(GRAPH_ROOT, filename + '.csv')
    png_filename = os.path.join(GRAPH_ROOT, filename + '.png')

    _check_directory_exists(csv_filename)
    _check_directory_exists(png_filename)

    if not settings.GRAPH_DEBUG or force_overwrite:
        if os.path.exists(csv_filename):
            if os.path.exists(png_filename):
                return

    institute_list = Institute.active.all()

    plt.subplots(figsize=(4, 4))

    data = []
    labels = []

    total = 0
    with open(csv_filename, 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        for institute in institute_list.iterator():
            hours, jobs = usage.get_institute_usage(institute, start, end)
            total = total + int(hours)
            if hours > 0:
                csv_writer.writerow([institute.name, hours, jobs])
                data.append(hours)
                labels.append(institute.name)

        mcu = usage.get_machine_category_usage(start, end)
        hours = int(mcu.available_time - total)
        csv_writer.writerow(["unused", hours])
        data.append(hours)
        labels.append('Unused')

    plt.pie(data, labels=labels, autopct='%1.1f%%', shadow=True)
    plt.tight_layout()
    plt.savefig(png_filename)
    plt.close()


def _gen_machine_graph(start, end, force_overwrite=False):
    """ Pie chart comparing machines usage. """
    filename = graphs.get_machine_graph_filename(start, end)
    csv_filename = os.path.join(GRAPH_ROOT, filename + '.csv')
    png_filename = os.path.join(GRAPH_ROOT, filename + '.png')

    _check_directory_exists(csv_filename)
    _check_directory_exists(png_filename)

    if not settings.GRAPH_DEBUG or force_overwrite:
        if os.path.exists(csv_filename):
            if os.path.exists(png_filename):
                return

    machine_list = Machine.objects.all()

    plt.subplots(figsize=(4, 4))

    data = []
    labels = []

    with open(csv_filename, 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        for machine in machine_list.iterator():
            hours, jobs = usage.get_machine_usage(machine, start, end)
            if hours > 0:
                csv_writer.writerow([machine.name, hours, jobs])
                data.append(hours)
                labels.append(machine.name)

    plt.pie(data, labels=labels, autopct='%1.1f%%', shadow=True)

    del data
    del labels

    plt.tight_layout()
    plt.savefig(png_filename)
    plt.close()


def _gen_trend_graph(start, end, force_overwrite=False):
    """ Total trend graph for machine category. """
    filename = graphs.get_trend_graph_filename(start, end)
    csv_filename = os.path.join(GRAPH_ROOT, filename + '.csv')
    png_filename = os.path.join(GRAPH_ROOT, filename + '.png')

    _check_directory_exists(csv_filename)
    _check_directory_exists(png_filename)

    if not settings.GRAPH_DEBUG or force_overwrite:
        if os.path.exists(csv_filename):
            if os.path.exists(png_filename):
                return

    query = CPUJob.objects.filter(
        date__range=(start, end)
    )
    query = query.values('date').annotate(Sum('cpu_usage'))
    query = query.order_by('date')

    t_start = start
    t_end = end

    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')

    fig, ax = plt.subplots(figsize=(6, 4))
    ax.set_xlim(start, end)
    ax.set_title('%s - %s' % (start_str, end_str))
    ax.set_ylabel("CPU Time (hours)")
    ax.set_xlabel("Date")

    locator = mdates.AutoDateLocator()
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(locator))
    ax.xaxis.set_minor_locator(mdates.DayLocator())

    data = {}
    x_data = []
    y_data = []

    with open(csv_filename, 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        for row in query.iterator():
            csv_writer.writerow([
                row['date'], row['cpu_usage__sum'] / 3600.00
            ])

            date = row['date']

            data[date] = row['cpu_usage__sum']

    start = t_start
    end = t_end
    while start <= end:
        total = 0
        if start in data:
            total = data[start]
        x_data.append(start)
        y_data.append(total / 3600.00)
        start = start + datetime.timedelta(days=1)

    del data

    ax.plot(x_data, y_data)

    del x_data
    del y_data

    fig.autofmt_xdate()
    plt.tight_layout()
    plt.savefig(png_filename)
    plt.close()


def _gen_institute_trend_graph(institute,
                               start,
                               end,
                               force_overwrite=False):
    """ Institute trend graph for machine category. """
    filename = graphs.get_institute_trend_graph_filename(institute, start, end)
    csv_filename = os.path.join(GRAPH_ROOT, filename + '.csv')
    png_filename = os.path.join(GRAPH_ROOT, filename + '.png')

    _check_directory_exists(csv_filename)
    _check_directory_exists(png_filename)

    if not settings.GRAPH_DEBUG or force_overwrite:
        if os.path.exists(csv_filename):
            if os.path.exists(png_filename):
                return

    query = CPUJob.objects.filter(
        project__institute=institute,
        date__range=(start, end)
    )
    query = query.values('date').annotate(Sum('cpu_usage'))
    query = query.order_by('date')

    t_start = start
    t_end = end

    start_str = start.strftime('%Y-%m-%d')
    end_str = end.strftime('%Y-%m-%d')

    fig, ax = plt.subplots(figsize=(6, 4))
    ax.set_xlim(start, end)
    ax.set_title('%s - %s' % (start_str, end_str))
    ax.set_ylabel("CPU Time (hours)")
    ax.set_xlabel("Date")

    locator = mdates.AutoDateLocator()
    ax.xaxis.set_major_locator(locator)
    ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(locator))
    ax.xaxis.set_minor_locator(mdates.DayLocator())

    data = {}
    x_data = []
    y_data = []

    with open(csv_filename, 'w') as csv_file:
        csv_writer = csv.writer(csv_file)
        for row in query.iterator():
            csv_writer.writerow([
                row['date'], row['cpu_usage__sum'] / 3600.00
            ])

            date = row['date']

            data[date] = row['cpu_usage__sum']

    start = t_start
    end = t_end
    while start <= end:
        total = 0
        if start in data:
            total = data[start]
        x_data.append(start)
        y_data.append(total / 3600.00)
        start = start + datetime.timedelta(days=1)

    del data

    ax.plot(x_data, y_data)

    del x_data
    del y_data

    fig.autofmt_xdate()
    plt.tight_layout()
    plt.savefig(png_filename)
    plt.close()



# -*- coding: utf-8 -*-
# Generated by Django 1.11.10 on 2018-03-02 18:08
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('kgapplications', '0004_auto_20180301_1721'),
    ]

    operations = [
        migrations.AlterField(
            model_name='application',
            name='created_by',
            field=models.ForeignKey(blank=True, editable=False, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-15 18:31
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('kgapplications', '0002_auto_20161007_1821'),
        ('karaage', '0004_auto_20160429_0927'),
    ]

    run_before = [
        ('karaage', '0005_auto_20171215_1831'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='projectapplication',
            name='machine_categories',
        ),
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import getpass

from django.core.management.base import BaseCommand, CommandError

from karaage.people.models import Person


class Command(BaseCommand):
    help = "Change a person's password for Karaage."

    def _get_pass(self, prompt="Password: "):
        p = getpass.getpass(prompt=prompt)
        if not p:
            raise CommandError("aborted")
        return p

    def handle(self, *args, **options):
        if len(args) > 1:
            raise CommandError(
                "need exactly one or zero arguments for username")

        if args:
            username, = args
        else:
            username = getpass.getuser()

        try:
            person = Person.objects.get(username=username)
        except Person.DoesNotExist:
            raise CommandError("person '%s' does not exist" % username)

        self.stdout.write("Changing password for person '%s'\n" % person)

        max_tries = 3
        count = 0
        p1, p2 = 1, 2  # To make them initially mismatch.
        while p1 != p2 and count < max_tries:
            p1 = self._get_pass()
            p2 = self._get_pass("Password (again): ")
            if p1 != p2:
                self.stdout.write(
                    "Passwords do not match. Please try again.\n")
                count = count + 1

        if count == max_tries:
            raise CommandError(
                "Aborting password change for user '%s' after %s attempts"
                % (username, count))

        person.set_password(p1)
        person.save()

        return "Password changed successfully for user '%s'" % person



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" Forms for Karaage use. """

import re

import six
from django import forms
from django.conf import settings

from karaage.common.models import COMMENT, LogEntry

from .passwords import assert_strong_password


def validate_password(username, password1, password2=None, old_password=None):
    # password1 is mandatory, it must be given in order to proceed.
    if password1 is None:
        raise forms.ValidationError(
            six.u("The first password was not given."))

    # password2 is an optional parameter, and may be None
    if password2 is not None and password1 != password2:
        raise forms.ValidationError(
            six.u("The two password fields didn't match."))

    # Now we can check password1
    try:
        assert_strong_password(username, password1, old_password)
    except ValueError as e:
        raise forms.ValidationError(six.u(
            'Your password was found to be insecure: %s.' % str(e)))

    # If password1 is ok, return it.
    return password1


def clean_email(email):
    email_match_type = "exclude"
    email_match_list = []
    if hasattr(settings, 'EMAIL_MATCH_TYPE'):
        email_match_type = settings.EMAIL_MATCH_TYPE
    if hasattr(settings, 'EMAIL_MATCH_LIST'):
        email_match_list = settings.EMAIL_MATCH_LIST

    found = False
    for string in email_match_list:
        m = re.search(string, email, re.IGNORECASE)
        if m is not None:
            found = True
            break

    message = "This email address cannot be used."
    if hasattr(settings, 'EMAIL_MATCH_MSG'):
        message = settings.EMAIL_MATCH_MSG
    if email_match_type == "include":
        if not found:
            raise forms.ValidationError(message)
    elif email_match_type == "exclude":
        if found:
            raise forms.ValidationError(message)
    else:
        raise forms.ValidationError("Oops. Nothing is valid. Sorry.")


class CommentForm(forms.ModelForm):
    """ Comment form. """

    class Meta:
        model = LogEntry
        fields = ["change_message"]

    def __init__(self, obj, request, instance=None, **kwargs):
        self.obj = obj
        self.request = request
        if instance is not None:
            assert instance.action_flag == COMMENT
            assert obj == instance.content_object
        super(CommentForm, self).__init__(instance=instance, **kwargs)

    def save(self, commit=True):
        log = super(CommentForm, self).save(commit=False)
        log.content_object = self.obj
        log.object_repr = six.text_type(self.obj)
        log.action_flag = COMMENT
        log.user = self.request.user

        if commit:
            log.save()

        return log


class LoginForm(forms.Form):
    username = forms.CharField(label="Username", max_length=30)
    password = forms.CharField(label="Password", widget=forms.PasswordInput)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import datetime

from django.db import models
from django.urls import reverse
from django.utils.encoding import python_2_unicode_compatible
from model_utils import FieldTracker

from karaage.common import is_admin, log
from karaage.institutes.models import Institute
from karaage.machines.models import Account
from karaage.people.models import Group, Person
from karaage.projects.managers import (
    ActiveProjectManager,
    DeletedProjectManager,
)


@python_2_unicode_compatible
class Project(models.Model):
    pid = models.CharField(max_length=255, unique=True)
    name = models.CharField(max_length=200)
    group = models.ForeignKey(Group, on_delete=models.PROTECT)
    institute = models.ForeignKey(Institute, on_delete=models.CASCADE)
    leaders = models.ManyToManyField(Person, related_name='leads')
    description = models.TextField(null=True, blank=True)
    is_approved = models.BooleanField(default=False)
    start_date = models.DateField(default=datetime.datetime.today)
    end_date = models.DateField(null=True, blank=True)
    additional_req = models.TextField(null=True, blank=True)
    is_active = models.BooleanField(default=False)
    approved_by = models.ForeignKey(
        Person, related_name='project_approver',
        null=True, blank=True, editable=False,
        on_delete=models.SET_NULL)
    date_approved = models.DateField(null=True, blank=True, editable=False)
    deleted_by = models.ForeignKey(
        Person, related_name='project_deletor',
        null=True, blank=True, editable=False,
        on_delete=models.SET_NULL)
    date_deleted = models.DateField(null=True, blank=True, editable=False)
    last_usage = models.DateField(null=True, blank=True, editable=False)
    objects = models.Manager()
    active = ActiveProjectManager()
    deleted = DeletedProjectManager()

    _tracker = FieldTracker()

    class Meta:
        ordering = ['pid']
        db_table = 'project'
        app_label = 'karaage'

    def __str__(self):
        return '%s - %s' % (self.pid, self.name)

    def get_absolute_url(self):
        return reverse('kg_project_detail', args=[self.id])

    def save(self, *args, **kwargs):
        created = self.pk is None

        # set group if not already set
        if self.group_id is None:
            name = self.pid
            self.group, _ = Group.objects.get_or_create(name=name)

        # save the object
        super(Project, self).save(*args, **kwargs)

        if created:
            log.add(self, 'Created')
        for field in self._tracker.changed():
            log.change(self, 'Changed %s to %s'
                       % (field, getattr(self, field)))

        # has pid changed?
        self._tracker.has_changed("pid")
        if self._tracker.has_changed("pid"):
            old_pid = self._tracker.previous('pid')
            if old_pid is not None:
                from karaage.datastores import set_project_pid
                set_project_pid(self, old_pid, self.pid)
                log.change(self, 'Renamed %s to %s' % (old_pid, self.pid))

        # update the datastore
        from karaage.datastores import save_project
        save_project(self)

        # has group changed?
        if self._tracker.has_changed("group_id"):
            old_group_pk = self._tracker.previous("group_id")
            new_group = self.group
            if old_group_pk is not None:
                old_group = Group.objects.get(pk=old_group_pk)
                from karaage.datastores import remove_accounts_from_project
                query = Account.objects.filter(person__groups=old_group)
                remove_accounts_from_project(query, self)
            if new_group is not None:
                from karaage.datastores import add_accounts_to_project
                query = Account.objects.filter(person__groups=new_group)
                add_accounts_to_project(query, self)
    save.alters_data = True

    def delete(self, *args, **kwargs):
        # ensure nothing connects to this project
        query = Account.objects.filter(default_project=self)
        query.update(default_project=None)

        # Get list of accounts.
        # We do this here to ensure nothing gets deleted when
        # we call the super method.
        old_group_pk = self._tracker.previous("group_id")
        if old_group_pk is not None:
            old_group = Group.objects.get(pk=old_group_pk)
            query = Account.objects.filter(person__groups=old_group)
            query = query.filter(date_deleted__isnull=True)
            accounts = list(query)
        else:
            accounts = []

        # delete the object
        log.delete(self, 'Deleted')
        super(Project, self).delete(*args, **kwargs)

        # update datastore associations
        for account in accounts:
            from karaage.datastores import remove_account_from_project
            remove_account_from_project(account, self)

        # update the datastore
        from karaage.datastores import delete_project
        delete_project(self)
    delete.alters_data = True

    # Can user view this self record?
    def can_view(self, request):
        person = request.user

        if not person.is_authenticated:
            return False

        if is_admin(request):
            return True

        if not self.is_active:
            return False

        if not person.is_active:
            return False

        if person.is_locked():
            return False

        # Institute delegate==person can view any member of institute
        if self.institute.is_active:
            if person in self.institute.delegates.all():
                return True

        # Leader==person can view projects they lead
        tmp = self.leaders.filter(pk=person.pk)
        if tmp.count() > 0:
            return True

        # person can view own projects
        tmp = self.group.members.filter(pk=person.pk)
        if tmp.count() > 0:
            return True

        return False

    # Can user view this self record?
    def can_edit(self, request):
        # The same as can_view, except normal project members cannot edit
        person = request.user

        if not person.is_authenticated:
            return False

        if is_admin(request):
            return True

        if not self.is_active:
            return False

        if not person.is_active:
            return False

        if person.is_locked():
            return False

        # Institute delegate==person can edit any member of institute
        if self.institute.is_active:
            if person in self.institute.delegates.all():
                return True

        # Leader==person can edit projects they lead
        tmp = self.leaders.filter(pk=person.pk)
        if tmp.count() > 0:
            return True

        return False

    def activate(self, approved_by):
        if self.is_active is True:
            return
        self.is_active = True
        self.is_approved = True
        self.date_approved = datetime.datetime.today()
        self.approved_by = approved_by
        self.save()
        log.change(self, 'Activated by %s' % approved_by)
    activate.alters_data = True

    def deactivate(self, deleted_by):
        self.is_active = False
        self.deleted_by = deleted_by
        self.date_deleted = datetime.datetime.today()
        self.group.members.clear()
        self.save()
        log.change(self, 'Deactivated by %s' % deleted_by)
    deactivate.alters_data = True


def _leaders_changed(
        sender, instance, action, reverse, model, pk_set, **kwargs):
    """
    Hook that executes whenever the group members are changed.
    """
    # print("'%s','%s','%s','%s','%s'"
    #   %(instance, action, reverse, model, pk_set))

    if action == "post_add":
        if not reverse:
            project = instance
            for person in model.objects.filter(pk__in=pk_set):
                log.change(
                    person,
                    "Added person to project %s leaders" % project)
                log.change(
                    project,
                    "Added person %s to project leaders" % person)
        else:
            person = instance
            for project in model.objects.filter(pk__in=pk_set):
                log.change(
                    person,
                    "Added person to project %s leaders" % project)
                log.change(
                    project,
                    "Added person %s to project leaders" % person)

    elif action == "post_remove":
        if not reverse:
            project = instance
            for person in model.objects.filter(pk__in=pk_set):
                log.change(
                    person,
                    "Removed person from project %s leaders" % project)
                log.change(
                    project,
                    "Removed person %s from project leaders" % person)
        else:
            person = instance
            for project in model.objects.filter(pk__in=pk_set):
                log.change(
                    person,
                    "Removed person from project %s leaders" % project)
                log.change(
                    project,
                    "Removed person %s from project leaders" % person)

    elif action == "pre_clear":
        # This has to occur in pre_clear, not post_clear, as otherwise
        # we won't see what project leaders need to be removed.
        if not reverse:
            project = instance
            log.change(
                project,
                "Removed all project leaders")
            for person in project.leaders.all():
                log.change(
                    project,
                    "Removed person %s from project leaders" % person)
        else:
            person = instance
            log.change(
                person,
                "Removed person from all project leaders")
            for project in person.leads.all():
                log.change(
                    project,
                    "Removed person %s from project leaders" % person)


models.signals.m2m_changed.connect(
    _leaders_changed, sender=Project.leaders.through)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.7 on 2017-08-31 17:35
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kgusage', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='cpujob',
            name='software',
            field=models.ManyToManyField(blank=True, to='kgsoftware.SoftwareVersion'),
        ),
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" This file shows the project application views using a state machine. """

from django.conf import settings
from django.http import HttpResponseForbidden, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse

from karaage.common import is_admin, saml
from karaage.common.decorators import login_required
from karaage.people.models import Person
from karaage.projects.models import Project

from .. import forms
from ..models import Applicant, ProjectApplication
from . import base


def get_application_state_machine():
    """ Get the default state machine for applications. """
    config = settings.APPLICATION_PROJECT
    state_machine = base.StateMachine(config)
    return state_machine


def register():
    base.setup_application_type(
        ProjectApplication, get_application_state_machine())


def get_applicant_from_email(email):
    """
    Get applicant from email address.

    If the person exists, return (person, True)

    If multiple matches, return (None, True)

    Otherwise create applicant and return (applicant, False)
    """
    try:
        applicant = Person.active.get(email=email)
        existing_person = True
    except Person.DoesNotExist:
        applicant, _ = Applicant.objects.get_or_create(email=email)
        existing_person = False
    except Person.MultipleObjectsReturned:
        applicant = None
        existing_person = False
    return applicant, existing_person


def _send_invitation(request, project):
    """ The logged in project leader OR administrator wants to invite somebody.
    """
    form = forms.InviteUserApplicationForm(request.POST or None)
    if request.method == 'POST':
        if form.is_valid():

            email = form.cleaned_data['email']
            applicant, existing_person = get_applicant_from_email(email)

            # If applicant is None then there were multiple persons found.
            if applicant is None:
                return render(
                    template_name='kgapplications/'
                                  'project_common_invite_multiple.html',
                    context={'form': form, 'email': email},
                    request=request)

            if existing_person and 'existing' not in request.POST:
                return render(
                    template_name='kgapplications/'
                                  'project_common_invite_existing.html',
                    context={'form': form, 'person': applicant},
                    request=request)

            application = form.save(commit=False)
            application.applicant = applicant
            application.project = project
            application.save()
            state_machine = get_application_state_machine()
            response = state_machine.start(request, application)
            return response

    return render(
        template_name='kgapplications/project_common_invite_other.html',
        context={'form': form, 'project': project, },
        request=request)


@login_required
def send_invitation(request, project_id=None):
    """ The logged in project leader wants to invite somebody to their project.
    """

    project = None
    if project_id is not None:
        project = get_object_or_404(Project, id=project_id)

    if project is None:

        if not is_admin(request):
            return HttpResponseForbidden('<h1>Access Denied</h1>')

    else:

        if not project.can_edit(request):
            return HttpResponseForbidden('<h1>Access Denied</h1>')

    return _send_invitation(request, project)


def new_application(request):
    """ A new application by a user to start a new project. """
    # Note default kgapplications/index.html will display error if user logged
    # in.
    if not settings.ALLOW_REGISTRATIONS:
        return render(
            template_name='kgapplications/project_common_disabled.html',
            context={},
            request=request)

    roles = {'is_applicant', 'is_authorised'}

    if not request.user.is_authenticated:
        attrs, _ = saml.parse_attributes(request)
        defaults = {'email': attrs['email']}
        form = forms.UnauthenticatedInviteUserApplicationForm(
            request.POST or None, initial=defaults)
        if request.method == 'POST':
            if form.is_valid():
                email = form.cleaned_data['email']
                applicant, existing_person = get_applicant_from_email(email)

                # If applicant is None then there were multiple persons found.
                # This should never happen as the
                # UnauthenticatedInviteUserApplicationForm form disallows
                # existing users applying unauthenticated.
                assert applicant is not None
                # Similarly existing_person should always be False here.
                assert not existing_person

                application = ProjectApplication()
                application.applicant = applicant
                application.save()

                state_machine = get_application_state_machine()
                state_machine.start(request, application, roles)
                # we do not show unauthenticated users the application at this
                # stage.
                url = reverse('index')
                return HttpResponseRedirect(url)
        return render(
            template_name='kgapplications/'
            'project_common_invite_unauthenticated.html',
            context={'form': form, },
            request=request)
    else:
        if request.method == 'POST':
            person = request.user

            application = ProjectApplication()
            application.applicant = person
            application.save()

            state_machine = get_application_state_machine()
            response = state_machine.start(request, application, roles)
            return response
        return render(
            template_name='kgapplications/'
            'project_common_invite_authenticated.html',
            context={},
            request=request)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import six
from django import forms
from django.conf import settings

from karaage.institutes.models import Institute


def is_saml_session(request):
    if not settings.SHIB_SUPPORTED:
        return False
    if 'HTTP_SHIB_SESSION_ID' not in request.META:
        return False
    if not request.META['HTTP_SHIB_SESSION_ID']:
        return False
    return True


def parse_attributes(request):
    shib_attrs = {}
    error = False
    for header, attr in settings.SHIB_ATTRIBUTE_MAP.items():
        required, name = attr
        values = request.META.get(header, None)
        value = None
        if values:
            # If multiple attributes releases just care about the 1st one
            try:
                value = values.split(';')[0]
            except IndexError:
                value = values

        shib_attrs[name] = value
        if not value or value == '':
            if required:
                error = True
    return shib_attrs, error


def build_shib_url(request, target, entityid=None):
    url_base = 'https://%s' % request.get_host()
    shib_url = "%s%s" % (
        url_base, getattr(
            settings, 'SHIB_HANDLER_LOGIN', '/Shibboleth.sso/Login'))
    if not target.startswith('http'):
        target = url_base + target

    url = '%s?target=%s' % (shib_url, target)
    if entityid:
        url += '&entityID=%s' % entityid
    return url


def logout_url(request):
    url_base = 'https://%s' % request.get_host()
    url = "%s%s" % (
        url_base, getattr(
            settings, 'SHIB_HANDLER_LOGOUT', '/Shibboleth.sso/Logout'))
    return url


def add_saml_data(person, request):
    attrs, error = parse_attributes(request)

    # fill name if it was supplied
    if 'first_name' in attrs and attrs['first_name'] is not None:
        person.short_name = attrs['first_name']
        if 'last_name' in attrs and attrs['last_name'] is not None:
            person.full_name = six.u("%s %s") % (
                attrs['first_name'], attrs['last_name'])

    # short_name and full_name cannot be None
    if person.short_name is None:
        person.short_name = ""
    if person.full_name is None:
        person.full_name = ""

    # fill uid if it was supplied
    person.username = attrs.get('uid', person.username) \
        or person.username

    # fill telephone if supplied
    person.telephone = attrs.get('telephone', person.telephone) \
        or person.telephone

    # fill in mandatory attributes
    person.email = attrs['email']
    person.saml_id = attrs['persistent_id']
    person.institute = Institute.objects.get(saml_entityid=attrs['idp'])
    person.email_verified = True

    # save person
    person.save()
    return person


class SAMLInstituteForm(forms.Form):
    institute = forms.ModelChoiceField(queryset=None, required=True)

    def __init__(self, *args, **kwargs):
        super(SAMLInstituteForm, self).__init__(*args, **kwargs)

        queryset = Institute.active.filter(saml_entityid__isnull=False)
        queryset = queryset.exclude(saml_entityid="")

        self.fields['institute'].queryset = queryset



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.
from __future__ import absolute_import

import errno
import grp
import logging
import os.path
import pwd


class FileHandler(logging.FileHandler):

    def __init__(self, filename, owner=None, **kwargs):
        if owner:
            if not os.path.exists(filename):
                open(filename, 'a').close()
            uid = pwd.getpwnam(owner[0]).pw_uid
            gid = grp.getgrnam(owner[1]).gr_gid
            try:
                os.chown(filename, uid, gid)
            except OSError as ex:
                if ex.errno != errno.EPERM:
                    raise
        super(FileHandler, self).__init__(filename=filename, **kwargs)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.contrib.auth import authenticate
from django.contrib.auth.models import BaseUserManager


class PersonManager(BaseUserManager):

    def authenticate(self, username, password):
        return authenticate(username=username, password=password)

    def get_queryset(self):
        return super(PersonManager, self).get_queryset().select_related()

    def _create_user(
            self, username, email, short_name, full_name,
            institute, password, is_admin, **extra_fields):
        """Creates a new active person. """

        # Create Person
        person = self.model(
            username=username, email=email,
            short_name=short_name, full_name=full_name,
            is_admin=is_admin,
            institute=institute,
            **extra_fields
        )
        person.set_password(password)
        person.save()
        return person

    def create_user(
            self, username, email, short_name, full_name,
            institute, password=None, **extra_fields):
        """ Creates a new ordinary person. """
        return self._create_user(
            username=username, email=email,
            short_name=short_name, full_name=full_name,
            institute=institute, password=password,
            is_admin=False, **extra_fields)

    def create_superuser(
            self, username, email, short_name, full_name,
            institute, password, **extra_fields):
        """ Creates a new person with super powers. """
        return self._create_user(
            username=username, email=email,
            institute=institute, password=password,
            short_name=short_name, full_name=full_name,
            is_admin=True, **extra_fields)


class ActivePersonManager(PersonManager):

    def get_queryset(self):
        return super(ActivePersonManager, self) \
            .get_queryset() \
            .select_related() \
            .filter(is_active=True, is_systemuser=False)


class DeletedPersonManager(PersonManager):

    def get_queryset(self):
        return super(DeletedPersonManager, self) \
            .get_queryset() \
            .filter(is_active=False)


class LeaderManager(PersonManager):

    def get_queryset(self):
        leader_ids = []
        query = super(LeaderManager, self).get_queryset()
        query = query.filter(is_active=True)
        for l in query:
            if l.is_leader():
                leader_ids.append(l.id)

        return self.filter(id__in=leader_ids)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import datetime
import re

import django
import pytest
from django.conf import settings
from django.contrib import auth
from django.core import mail
from django.urls import reverse
from tldap.exceptions import ObjectDoesNotExist

from karaage.institutes.models import Institute, InstituteDelegate
from karaage.machines.models import Account
from karaage.people.models import Person
from karaage.projects.models import Project
from karaage.tests.integration import IntegrationTestCase


class FakeRequest(object):

    def __init__(self, person):
        self.user = person


@pytest.mark.django_db
class PersonTestCase(IntegrationTestCase):
    fixtures = [
        'test_karaage.json',
    ]

    def test_login(self):
        if django.VERSION >= (1, 9):
            url_prefix = ""
        else:
            url_prefix = 'http://testserver'

        form_data = {
            'username': 'kgsuper',
            'password': 'aq12ws',
        }

        response = self.client.post(
            reverse('kg_profile_login'), form_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.redirect_chain[0][0],
            url_prefix + reverse('index'))

        user = auth.get_user(self.client)
        assert user.is_authenticated
        assert user.username == 'kgsuper'

    def test_saml_login(self):
        person = Person.objects.get(username='kgtestuser1')
        person.institute = Institute.objects.get(pk=3)
        person.saml_id = "pY9RL3RHeQwAm4Hd"
        person.save()

        form_data = {
            'login': 'true',
            'institute': 3,
        }

        response = self.client.post(
            reverse('kg_profile_login_saml'), form_data, follow=True)
        self.assertEqual(response.status_code, 404)
        self.assertEqual(
            response.redirect_chain[0][0],
            "https://testserver/Shibboleth.sso/Login"
            + "?target=https://testserver/&entityID=http://samlid")

        response = self.client.get(
            reverse('index'),
            HTTP_SHIB_SESSION_ID="yes",
            HTTP_SHIB_IDENTITY_PROVIDER="http://samlid",
            HTTP_PERSISTENT_ID="pY9RL3RHeQwAm4Hd",
            HTTP_MAIL="email@example.org",
            HTTP_GIVENNAME="Firstname",
            HTTP_SN="Surname",
            HTTP_TELEPHONENUMBER="000",
            HTTP_UID="somethingelse"
        )
        self.assertEqual(response.status_code, 200)

        user = auth.get_user(self.client)
        assert user.is_authenticated
        assert user.username == 'kgtestuser1'

    def test_logout(self):
        if django.VERSION >= (1, 9):
            url_prefix = ""
        else:
            url_prefix = 'http://testserver'

        response = self.client.post(reverse('kg_profile_logout'), follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.redirect_chain[0][0],
            url_prefix + reverse('index'))

        user = auth.get_user(self.client)
        assert not user.is_authenticated

    def do_permission_tests(self, test_object, users):
        for user_id in users:
            # print("can user '%d' access '%s'?"%(user_id, test_object))
            person = Person.objects.get(id=user_id)
            request = FakeRequest(person)
            result = test_object.can_view(request)
            expected_result = users[user_id]
            # print("---> got:'%s' expected:'%s'"%(result, expected_result))
            self.assertEqual(
                result, expected_result,
                "%r.can_view(%r) returned %r but we expected %r"
                % (test_object, person, result, expected_result))
            # print()

    def test_permissions(self):
        test_object = Project.objects.get(pid="TestProject1")
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate,
                        #                    project leader
            2: False,   # person 2 cannot view
            3: True,    # person 3 can view: project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=1)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: self, project member,
                        #                    person's institute delegate
            2: False,   # person 2 cannot view
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: is_staff, institute delegate
        })

        test_object = Person.objects.get(id=2)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate
            2: True,    # person 2 can view: self
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=3)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate,
                        # project leader
            2: False,   # person 2 cannot view
            3: True,    # person 3 can view: self, project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=4)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate
            2: False,   # person 2 cannot view
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: self, is_staff
        })

        # add user 2 to project
        # test that members can see other people in own project
        # print("------------------------------------------------------------")
        project = Project.objects.get(pid="TestProject1")
        project.group.members.set([2, 3])

        test_object = Project.objects.get(pid="TestProject1")
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate
            2: True,    # person 2 can view: project member
            3: True,    # person 3 can view: project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=1)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: self, project member,
                        #                    delegate of institute
            2: False,   # person 2 cannot view
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=2)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate,
                        #                    project leader
            2: True,    # person 2 can view: self
            3: True,    # person 3 can view: project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=3)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate,
                        #                    project leader
            2: True,    # person 2 can view: project member
            3: True,    # person 3 can view: self, project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=4)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate
            2: False,   # person 2 cannot view
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: self, is_staff
        })

        # change institute of all people
        # Test institute leader can access people in project despite not being
        # institute leader for these people.
        # print("------------------------------------------------------------")
        Person.objects.all().update(institute=2)
        # Institute.objects.filter(pk=2).update(delegate=2,active_delegate=2)
        InstituteDelegate.objects.get_or_create(
            institute=Institute.objects.get(id=2),
            person=Person.objects.get(id=2),
            defaults={'send_email': False})
        project = Project.objects.get(pid="TestProject1")
        project.leaders.set([2])

        test_object = Project.objects.get(pid="TestProject1")
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate
            2: True,    # person 2 can view: project member, person's
                        #                    institute delegate, project leader
            3: True,    # person 3 can view: project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=1)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: self, project member
            2: True,    # person 2 can view: person's institute delegate
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=2)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: project's institute leader
            2: True,    # person 2 can view: self, person's institute delegate,
                        #                    project leader
            3: True,    # person 3 can view: project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=3)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: project's institute leader
            2: True,    # person 2 can view: project member, person's institute
                        #                    delegate, project leader
            3: True,    # person 3 can view: self, project member
            4: True,    # person 4 can view: is_staff
        })

        test_object = Person.objects.get(id=4)
        self.do_permission_tests(test_object, {
            1: True,    # person 1 can view: person's institute delegate
            2: True,    # person 2 can view: person's institute delegate
            3: False,   # person 3 cannot view
            4: True,    # person 4 can view: self, is_staff
        })

    def test_admin_create_user_with_account(self):
        users = Person.objects.count()
        project = Project.objects.get(pid='TestProject1')
        p_users = project.group.members.count()
        logged_in = self.client.login(username='kgsuper', password='aq12ws')
        self.assertEqual(logged_in, True)
        response = self.client.get(reverse('kg_person_add'))
        self.assertEqual(response.status_code, 200)

        form_data = {
            'title': 'Mr',
            'short_name': 'Sam',
            'full_name': 'Sam Morrison',
            'position': 'Sys Admin',
            'institute': 1,
            'department': 'eddf',
            'email': 'sam2@vpac.org',
            'country': 'AU',
            'telephone': '4444444',
            'username': 'samtest',
            'password1': 'Exaiquouxei0',
            'password2': 'Exaiquouxei0',
            'project': 1,
            'needs_account': True,
        }

        response = self.client.post(reverse('kg_person_add'), form_data)
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Person.objects.count(), users + 1)
        users = users + 1
        person = Person.objects.get(username='samtest')
        self.assertEqual(person.is_active, True)
        self.assertEqual(person.username, 'samtest')
        self.assertEqual(Account.objects.count(), 2)
        self.assertEqual(project.group.members.count(), p_users + 1)
        luser = self._ldap_datastore._get_account(uid='samtest')
        self.assertEqual(luser['givenName'], ['Sam'])
        self.assertEqual(luser['homeDirectory'], ['/vpac/TestProject1/samtest'])

    def test_admin_create_user(self):
        users = Person.objects.count()
        project = Project.objects.get(pid='TestProject1')
        project.group.members.count()
        logged_in = self.client.login(username='kgsuper', password='aq12ws')
        self.assertEqual(logged_in, True)
        response = self.client.get(reverse('kg_person_add'))

        self.assertEqual(response.status_code, 200)

        form_data = {
            'title': 'Mr',
            'short_name': 'Sam',
            'full_name': 'Sam Morrison2',
            'position': 'Sys Admin',
            'institute': 1,
            'department': 'eddf',
            'email': 'sam2@vpac.org',
            'country': 'AU',
            'telephone': '4444444',
            'username': 'samtest2',
            'password1': 'Exaiquouxei0',
            'password2': 'Exaiquouxei0',
            'needs_account': False,
        }

        response = self.client.post(reverse('kg_person_add'), form_data)
        self.assertEqual(response.status_code, 302)

        self.assertEqual(Person.objects.count(), users + 1)
        person = Person.objects.get(username='samtest2')
        self.assertEqual(person.is_active, True)
        self.assertEqual(person.username, 'samtest2')
        # Try adding it again - Should fail
        response = self.client.post(reverse('kg_person_add'), form_data)
        self.assertEqual(response.status_code, 200)

    def test_admin_update_person(self):
        logged_in = self.client.login(username='kgsuper', password='aq12ws')
        self.assertEqual(logged_in, True)

        person = Person.objects.get(username='kgtestuser3')
        self.assertEqual(person.mobile, '')
        luser = self._ldap_datastore._get_account(uid='kgtestuser3')
        self.assertEqual(luser['gidNumber'], [500])
        self.assertEqual(luser['o'], ['Example'])
        self.assertEqual(luser['gecos'], ['Test User3 (Example)'])
        response = self.client.get(
            reverse('kg_person_edit', args=['kgtestuser3']))
        self.assertEqual(response.status_code, 200)

        form_data = {
            'title': 'Mr',
            'short_name': 'Test',
            'full_name': 'Test User3',
            'position': 'Sys Admin',
            'institute': 2,
            'department': 'eddf',
            'email': 'sam2@vpac.org',
            'country': 'AU',
            'telephone': '4444444',
            'mobile': '555666',
        }
        response = self.client.post(
            reverse('kg_person_edit', args=['kgtestuser3']), form_data)
        self.assertEqual(response.status_code, 302)

        person = Person.objects.get(username='kgtestuser3')
        self.assertEqual(person.mobile, '555666')
        luser = self._ldap_datastore._get_account(uid='kgtestuser3')
        self.assertEqual(luser['gidNumber'], [501])
        self.assertEqual(luser['o'], ['OtherInst'])
        self.assertEqual(luser['gecos'], ['Test User3 (OtherInst)'])

    def test_delete_activate_person(self):
        self.client.login(username='kgsuper', password='aq12ws')
        person = Person.objects.get(username='kgtestuser3')
        self.assertEqual(person.is_active, True)
        self.assertEqual(person.projects.count(), 1)
        self.assertEqual(person.account_set.count(), 1)
        self.assertEqual(person.account_set.all()[0].date_deleted, None)
        luser = self._ldap_datastore._get_account(uid='kgtestuser3')
        self.assertEqual(luser['givenName'], ['Test'])

        response = self.client.get(
            reverse('kg_person_delete', args=[person.username]))
        self.assertEqual(response.status_code, 200)
        # Test deleting
        response = self.client.post(
            reverse('kg_person_delete', args=[person.username]))
        self.assertEqual(response.status_code, 302)

        person = Person.objects.get(username='kgtestuser3')

        self.assertEqual(person.is_active, False)
        self.assertEqual(person.projects.count(), 0)
        self.assertEqual(person.account_set.count(), 1)
        self.assertEqual(person.account_set.all()[0].date_deleted,
                         datetime.date.today())
        self.assertRaises(
            ObjectDoesNotExist,
            self._ldap_datastore._get_account,
            uid='kgtestuser3')

        # Test activating
        response = self.client.post(
            reverse('kg_person_activate', args=[person.username]))
        self.assertEqual(response.status_code, 302)
        person = Person.objects.get(username='kgtestuser3')
        self.assertEqual(person.is_active, True)

    def stest_delete_account(self):
        person = Person.objects.get(pk=Person.objects.count())
        ua = person.account_set.all()[0]
        self.assertEqual(person.is_active, True)
        self.assertEqual(person.account_set.count(), 1)
        self.assertEqual(ua.date_deleted, None)

        response = self.client.post(
            '/%susers/accounts/delete/%i/' % (settings.BASE_URL, ua.id))
        self.assertEqual(response.status_code, 302)

        person = Person.objects.get(pk=Person.objects.count())
        ua = person.account_set.all()[0]
        self.assertEqual(ua.date_deleted, datetime.date.today())
        self.assertEqual(person.project_set.count(), 0)

    def stest_default_projects(self):

        person = Person.objects.get(pk=Person.objects.count())
        ua = person.account_set.all()[0]

        self.assertEqual(person.project_set.count(), 1)
        self.assertEqual(person.project_set.all()[0], ua.default_project)
        project = Project.objects.create(
            pid='test2',
            name='test project',
            leader=person,
            start_date=datetime.date.today(),
            institute=Institute.objects.get(name='VPAC'),
            is_active=True,
            is_approved=True,
        )
        project.users.add(person)
        self.assertEqual(person.project_set.count(), 2)
        # change default
        response = self.client.post(
            reverse('kg_account_set_default', args=[ua.id, project.pid]))

        self.assertEqual(response.status_code, 302)

        person = Person.objects.get(pk=Person.objects.count())
        ua = person.account_set.all()[0]
        project = Project.objects.get(pid='test2')

        self.assertEqual(person.project_set.count(), 2)
        self.assertEqual(project, ua.default_project)

    def stest_add_user_to_project(self):

        person = Person.objects.get(pk=Person.objects.count())
        assert len(person.account_set.all()) == 1

        self.assertEqual(person.project_set.count(), 1)

        Project.objects.create(
            pid='test2',
            name='test project 5',
            leader=Person.objects.get(username='leader'),
            start_date=datetime.date.today(),
            institute=Institute.objects.get(name='VPAC'),
            is_active=True,
            is_approved=True,
        )

        response = self.client.post(
            reverse('kg_person_detail', args=[person.username]),
            {'project': 'test2', 'project-add': 'true'})
        self.assertEqual(response.status_code, 200)
        self.assertEqual(person.project_set.count(), 2)

    def test_password_reset_by_self(self):
        logged_in = self.client.login(
            username='kgtestuser1', password='aq12ws')
        self.assertEqual(logged_in, True)

        if django.VERSION >= (1, 9):
            url_prefix = ""
        else:
            url_prefix = 'http://testserver'

        # send request
        url = reverse("kg_profile_reset")
        done_url = reverse("kg_profile_reset_done")
        response = self.client.post(url, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.redirect_chain[0][0],
                         url_prefix + done_url)

        # check email
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, "TestOrg Password change")
        url = re.search(r"(?P<url>https?://[^\s]+)", message.body).group("url")
        self.assertTrue(
            url.startswith("https://example.com/users/persons/reset/"))
        url = url[25:]

        # get password reset page
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        # send new password
        form_data = {
            'new_password1': 'VQA#y!xD=BpI<sM69`RW:%',
            'new_password2': 'VQA#y!xD=BpI<sM69`RW:%',
        }
        done_url = reverse("password_reset_complete")
        response = self.client.post(url, form_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.redirect_chain[0][0],
                         url_prefix + done_url)

        # test new password
        logged_in = self.client.login(
            username='kgtestuser1', password='VQA#y!xD=BpI<sM69`RW:%')
        self.assertEqual(logged_in, True)

    def test_password_reset_by_admin(self):
        logged_in = self.client.login(username='kgsuper', password='aq12ws')
        self.assertEqual(logged_in, True)

        if django.VERSION >= (1, 9):
            url_prefix = ""
        else:
            url_prefix = 'http://testserver'

        # send request
        url = reverse("kg_person_reset", args=["kgtestuser1"])
        done_url = reverse("kg_person_reset_done", args=["kgtestuser1"])
        response = self.client.post(url, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.redirect_chain[0][0],
                         url_prefix + done_url)
        self.client.logout()

        # check email
        self.assertEqual(len(mail.outbox), 1)
        message = mail.outbox[0]
        self.assertEqual(message.subject, "TestOrg Password change")
        url = re.search(r"(?P<url>https?://[^\s]+)", message.body).group("url")
        self.assertTrue(
            url.startswith("https://example.com/users/persons/reset/"))
        url = url[25:]

        # get password reset page
        response = self.client.get(url)
        self.assertEqual(response.status_code, 200)

        # send new password
        form_data = {
            'new_password1': 'VQA#y!xD=BpI<sM69`RW:%',
            'new_password2': 'VQA#y!xD=BpI<sM69`RW:%',
        }
        done_url = reverse("password_reset_complete")
        response = self.client.post(url, form_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.redirect_chain[0][0],
                         url_prefix + done_url)

        # test new password
        logged_in = self.client.login(
            username='kgtestuser1', password='VQA#y!xD=BpI<sM69`RW:%')
        self.assertEqual(logged_in, True)



# -*- coding: utf-8 -*-
# Generated by Django 1.10.1 on 2016-10-07 18:21
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('kgapplications', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='applicant',
            name='email',
            field=models.EmailField(max_length=254),
        ),
        migrations.AlterField(
            model_name='applicant',
            name='username',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='projectapplication',
            name='machine_categories',
            field=models.ManyToManyField(blank=True, to='karaage.MachineCategory'),
        ),
    ]



# Copyright 2018, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from typing import Dict

import tldap.django.helpers as dhelpers
import tldap.fields
from tldap.database import (
    Changeset,
    Database,
    LdapObject,
    SearchOptions,
    helpers,
)


class BaseAccount(LdapObject):
    """ An abstract Generic LDAP account. """

    @classmethod
    def get_fields(cls) -> Dict[str, tldap.fields.Field]:
        fields = {
            **helpers.get_fields_common(),
            **helpers.get_fields_person(),
            **helpers.get_fields_account(),
            **helpers.get_fields_shadow(),
            'default_project': tldap.fields.FakeField(required=True),
        }
        return fields

    @classmethod
    def get_search_options(cls, database: Database) -> SearchOptions:
        settings = database.settings
        return SearchOptions(
            base_dn=settings['LDAP_ACCOUNT_BASE'],
            object_class={'inetOrgPerson', 'organizationalPerson', 'person'},
            pk_field="uid",
        )

    @classmethod
    def on_load(cls, python_data: LdapObject, _database: Database) -> LdapObject:
        python_data = helpers.load_person(python_data, OpenldapGroup)
        python_data = helpers.load_account(python_data, OpenldapGroup)
        python_data = helpers.load_shadow(python_data)
        return python_data

    @classmethod
    def on_save(cls, changes: Changeset, database: Database) -> Changeset:
        settings = database.settings
        changes = helpers.save_person(changes, database)
        changes = helpers.save_account(changes, database, {'uid', 'cn', 'givenName', 'sn', 'o', 'default_project'})
        changes = helpers.save_shadow(changes)
        changes = helpers.rdn_to_dn(changes, 'uid', settings['LDAP_ACCOUNT_BASE'])
        return changes

    def __repr__(self):
        return f"user:{self['uid'][0]}"


class BaseGroup(LdapObject):
    """ An abstract generic LDAP Group. """

    @classmethod
    def get_fields(cls) -> Dict[str, tldap.fields.Field]:
        fields = {
            **helpers.get_fields_common(),
            **helpers.get_fields_group(),
        }
        return fields

    @classmethod
    def get_search_options(cls, database: Database) -> SearchOptions:
        settings = database.settings
        return SearchOptions(
            base_dn=settings['LDAP_GROUP_BASE'],
            object_class={'posixGroup'},
            pk_field="cn",
        )

    @classmethod
    def on_load(cls, python_data: LdapObject, _database: Database) -> LdapObject:
        return python_data

    @classmethod
    def on_save(cls, changes: Changeset, database: Database) -> Changeset:
        settings = database.settings
        changes = helpers.save_group(changes)
        changes = helpers.set_object_class(changes, ['top', 'posixGroup'])
        changes = helpers.rdn_to_dn(changes, 'cn', settings['LDAP_GROUP_BASE'])
        return changes

    @classmethod
    def add_member(cls, changes: Changeset, member: BaseAccount) -> Changeset:
        return helpers.add_group_member(changes, member)

    @classmethod
    def remove_member(cls, changes: Changeset, member: BaseAccount) -> Changeset:
        return helpers.remove_group_member(changes, member)

    def __repr__(self):
        return f"group:{self['cn'][0]}"


class OpenldapAccount(BaseAccount):
    """ An OpenLDAP specific account with the pwdpolicy schema. """

    @classmethod
    def get_fields(cls) -> Dict[str, tldap.fields.Field]:
        fields = {
            **BaseAccount.get_fields(),
            **helpers.get_fields_pwdpolicy(),
        }
        return fields

    @classmethod
    def on_load(cls, python_data: LdapObject, database: Database) -> LdapObject:
        python_data = BaseAccount.on_load(python_data, database)
        python_data = helpers.load_pwdpolicy(python_data)
        return python_data

    @classmethod
    def on_save(cls, changes: Changeset, database: Database) -> Changeset:
        changes = BaseAccount.on_save(changes, database)
        changes = dhelpers.save_account(changes, OpenldapAccount, database)
        changes = helpers.save_pwdpolicy(changes)
        changes = helpers.set_object_class(changes, ['top', 'person', 'inetOrgPerson', 'organizationalPerson',
                                                     'shadowAccount', 'posixAccount', 'pwdPolicy'])
        return changes


class OpenldapGroup(BaseGroup):
    """ An OpenLDAP specific group. """

    @classmethod
    def on_load(cls, python_data: LdapObject, database: Database) -> LdapObject:
        python_data = BaseGroup.on_load(python_data, database)
        python_data = helpers.load_group(python_data, OpenldapAccount)
        return python_data

    @classmethod
    def on_save(cls, changes: Changeset, database: Database) -> Changeset:
        changes = BaseGroup.on_save(changes, database)
        changes = dhelpers.save_group(changes, OpenldapGroup, database)
        return changes


class Ds389Account(BaseAccount):
    """ A DS389 specific account with the password object schema. """

    @classmethod
    def get_fields(cls) -> Dict[str, tldap.fields.Field]:
        fields = {
            **BaseAccount.get_fields(),
            **helpers.get_fields_password_object(),
        }
        return fields

    @classmethod
    def on_load(cls, python_data: LdapObject, database: Database) -> LdapObject:
        python_data = BaseAccount.on_load(python_data, database)
        python_data = helpers.load_password_object(python_data)
        return python_data

    @classmethod
    def on_save(cls, changes: Changeset, database: Database) -> Changeset:
        changes = BaseAccount.on_save(changes, database)
        changes = dhelpers.save_account(changes, Ds389Account, database)
        changes = helpers.save_password_object(changes)
        changes = helpers.set_object_class(changes, ['top', 'person', 'inetOrgPerson', 'organizationalPerson',
                                                     'shadowAccount', 'posixAccount', 'passwordObject'])
        return changes


class Ds389Group(BaseGroup):
    """ A DS389 specific group. """

    @classmethod
    def on_load(cls, python_data: LdapObject, database: Database) -> LdapObject:
        python_data = BaseGroup.on_load(python_data, database)
        python_data = helpers.load_group(python_data, Ds389Account)
        return python_data

    @classmethod
    def on_save(cls, changes: Changeset, database: Database) -> Changeset:
        changes = BaseGroup.on_save(changes, database)
        changes = dhelpers.save_group(changes, Ds389Group, database)
        return changes



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django.db.transaction
import tldap.transaction
from django.core.management.base import BaseCommand


class Command(BaseCommand):
    help = "Cleans up usage cache"

    @django.db.transaction.atomic
    @tldap.transaction.commit_on_success
    def handle(self, **options):
        verbose = int(options.get('verbosity'))
        from ... import models
        if verbose > 1:
            print("Clearing project cache")
        models.ProjectCache.objects.all().delete()
        if verbose > 1:
            print("Clearing institute cache")
        models.InstituteCache.objects.all().delete()
        if verbose > 1:
            print("Clearing user cache")
        models.PersonCache.objects.all().delete()
        if verbose > 1:
            print("Clearing machine cache")
        models.MachineCache.objects.all().delete()
        if verbose > 1:
            print("Clearing machine category cache")
        models.MachineCategoryCache.objects.all().delete()



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import os.path

from django.conf import settings


try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

GRAPH_TMP = os.path.join(settings.TMP_DIR, settings.GRAPH_TMP)
GRAPH_ROOT = os.path.join(settings.FILES_DIR, settings.GRAPH_DIR)
GRAPH_URL = urlparse.urljoin(settings.FILES_URL, settings.GRAPH_DIR)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django_filters
import django_tables2 as tables
from django.utils.safestring import mark_safe
from django_tables2.utils import A

from .models import LogEntry


class LogEntryFilter(django_filters.FilterSet):
    begin_action_time = django_filters.DateFilter(
        field_name="action_time", lookup_expr="gte")
    end_action_time = django_filters.DateFilter(
        field_name="action_time", lookup_expr="lte")

    class Meta:
        model = LogEntry
        fields = ("begin_action_time", "end_action_time",
                  "content_type", "object_id")


class LogEntryTable(tables.Table):
    user = tables.LinkColumn('kg_person_detail', args=[A('user.username')])
    content_object = tables.Column(
        linkify=True,
        verbose_name="Object",
        order_by=['content_type', 'object_id']
    )

    def render_action_flag(self, record):
        if record.is_addition():
            html = '<span class="addlink">Add</span>'
        elif record.is_change():
            html = '<span class="changelink">Change</span>'
        elif record.is_deletion():
            html = '<span class="deletelink">Delete</span>'
        elif record.is_comment():
            html = '<span class="commentlink">Comment</span>'
        else:
            html = '-'
        return mark_safe(html)

    class Meta:
        model = LogEntry
        fields = ('action_time', 'user', 'obj', 'action_flag',
                  'change_message')
        empty_text = "No items"



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import messages
from django.http import (
    HttpResponseBadRequest,
    HttpResponseForbidden,
    HttpResponseRedirect,
)
from django.shortcuts import get_object_or_404, render
from django.urls import reverse

import karaage.common as common
from karaage.common.decorators import admin_required, login_required
from karaage.machines.forms import (
    AddProjectForm,
    AdminAccountForm,
    UserAccountForm,
)
from karaage.machines.models import Account
from karaage.people.models import Person
from karaage.people.views.persons import user_list
from karaage.projects.models import Project
from karaage.projects.utils import add_user_to_project


@login_required
def profile_accounts(request):
    person = request.user
    accounts = person.account_set.filter(date_deleted__isnull=True)
    return render(
        template_name='karaage/machines/profile_accounts.html',
        context=locals(),
        request=request)


@admin_required
def add_account(request, username=None):
    person = get_object_or_404(Person, username=username)
    account = None

    form = AdminAccountForm(
        data=request.POST or None,
        instance=account, person=person, initial={'username': username})

    if request.method == 'POST':
        if form.is_valid():
            account = form.save()
            person = account.person
            return HttpResponseRedirect(person.get_absolute_url())

    return render(
        template_name='karaage/machines/account_form.html',
        context={'form': form, 'person': person, 'account': account},
        request=request)


@login_required
def account_detail(request, account_id):
    account = get_object_or_404(Account, pk=account_id)

    if not account.can_view(request):
        return HttpResponseForbidden(
            '<h1>Access Denied</h1>'
            '<p>You do not have permission to view details '
            'about this account.</p>')

    return render(
        template_name='karaage/machines/account_detail.html',
        context={'account': account, 'can_edit': account.can_edit(request)},
        request=request)


@login_required
def edit_account(request, account_id):
    account = get_object_or_404(Account, pk=account_id)

    if not account.can_edit(request):
        return HttpResponseForbidden(
            '<h1>Access Denied</h1>'
            '<p>You do not have permission to edit details '
            'of this account.</p>')

    if common.is_admin(request):
        person = account.person
        username = account.username
        form = AdminAccountForm(
            data=request.POST or None,
            instance=account, person=person, initial={'username': username})
    else:
        person = request.user
        assert account.person == person
        form = UserAccountForm(
            data=request.POST or None, instance=account)

    if request.method == 'POST':
        if form.is_valid():
            account = form.save()
            person = account.person
            return HttpResponseRedirect(person.get_absolute_url())

    return render(
        template_name='karaage/machines/account_form.html',
        context={'form': form, 'person': person, 'account': account},
        request=request)


@admin_required
def add_project(request, username):
    person = get_object_or_404(Person, username=username)

    # Add to project form
    form = AddProjectForm(request.POST or None)
    if request.method == 'POST':
        # Post means adding this user to a project
        if form.is_valid():
            project = form.cleaned_data['project']
            add_user_to_project(person, project)
            messages.success(
                request,
                "User '%s' was added to %s succesfully" % (person, project))
            return HttpResponseRedirect(person.get_absolute_url())

    return render(
        template_name='karaage/machines/person_add_project.html',
        context=locals(),
        request=request)


@admin_required
def delete_account(request, account_id):

    account = get_object_or_404(Account, pk=account_id)

    if request.method == 'POST':
        account.deactivate()
        messages.success(
            request,
            "User account for '%s' deleted succesfully" % account.person)
        return HttpResponseRedirect(account.get_absolute_url())

    return render(
        template_name='karaage/machines/account_confirm_delete.html',
        context=locals(),
        request=request)


@admin_required
def no_project_list(request):
    persons = Person.active.filter(
        groups__project__isnull=True, account__isnull=False)
    return user_list(request, persons, 'No projects')


@admin_required
def no_default_list(request):
    persons = Person.objects.filter(
        account__isnull=False,
        account__default_project__isnull=True,
        account__date_deleted__isnull=True)
    return user_list(request, persons, 'No default projects')


@admin_required
def no_account_list(request):
    person_id_list = []

    for u in Person.objects.all():
        if not u.has_account():
            person_id_list.append(u.id)

    persons = Person.objects.filter(id__in=person_id_list)
    return user_list(request, persons, 'No accounts')


@admin_required
def wrong_default_list(request):
    wrong = []
    for u in Person.active.all():
        for ua in u.account_set.filter(date_deleted__isnull=True):
            d = False
            for p in ua.project_list():
                if p == ua.default_project:
                    d = True
            if not d:
                if not u.is_locked():
                    wrong.append(u.id)

    persons = Person.objects.filter(id__in=wrong)
    return user_list(request, persons, 'Wrong default projects')


@login_required
def make_default(request, account_id, project_id):
    account = get_object_or_404(Account, pk=account_id)
    redirect = reverse('kg_account_detail', args=[account.pk])

    if not account.can_edit(request):
        return HttpResponseForbidden(
            '<h1>Access Denied</h1>'
            '<p>You do not have permission to edit details '
            'of this account.</p>')

    try:
        project = account.person.projects.get(pid=project_id)
    except Project.DoesNotExist:
        return HttpResponseForbidden(
            '<h1>Access Denied</h1>'
            '<p>Person owning account is not in this project.</p>')

    if request.method != 'POST':
        return HttpResponseBadRequest("<h1>Bad Request</h1>")

    account.default_project = project
    account.save()
    messages.success(request, "Default project changed succesfully")

    return HttpResponseRedirect(redirect)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.
import os
from unittest import skipUnless

import pkg_resources
import tldap.database
import tldap.transaction
from django.test import TestCase

from karaage.datastores import _DATASTORES
from karaage.datastores.ldap import DataStore
from karaage.middleware.threadlocals import reset


def skip_if_missing_requirements(*requirements):
    try:
        pkg_resources.require(*requirements)
        msg = ''
    except pkg_resources.DistributionNotFound:
        msg = 'Missing one or more requirements (%s)' % '|'.join(requirements)
    return skipUnless(msg == '', msg)


class IntegrationTestCase(TestCase):
    LDAP_CONFIG = {
        'DESCRIPTION': 'LDAP datastore',
        'ENGINE': 'karaage.datastores.ldap.AccountDataStore',
        'LDAP': 'default',
        'PRIMARY_GROUP': "institute",
        'DEFAULT_PRIMARY_GROUP': "dummy",
        'NUMBER_SCHEME': 'default',
        'LDAP_ACCOUNT_BASE': os.environ['LDAP_ACCOUNT_BASE'],
        'LDAP_GROUP_BASE': os.environ['LDAP_GROUP_BASE'],
        'HOME_DIRECTORY_FORMAT': "/vpac/{default_project}/{uid}",
        'GECOS_FORMAT': '{cn} ({o})',
    }

    def setUp(self):
        super(IntegrationTestCase, self).setUp()

        tldap.transaction.enter_transaction_management()

        self.addCleanup(self.cleanup)

        if os.environ['LDAP_TYPE'] == "openldap":
            config = {
                **self.LDAP_CONFIG,
                'ACCOUNT': 'karaage.datastores.ldap_schemas.OpenldapAccount',
                'GROUP': 'karaage.datastores.ldap_schemas.OpenldapGroup',

            }
        elif os.environ['LDAP_TYPE'] == "ds389":
            config = {
                **self.LDAP_CONFIG,
                'ACCOUNT': 'karaage.datastores.ldap_schemas.Ds389Account',
                'GROUP': 'karaage.datastores.ldap_schemas.Ds389Group',

            }
        else:
            raise RuntimeError(f"Unknown databasebase type {os.environ['LDAP_TYPE']}.")

        self._ldap_datastore = DataStore(config)

        database = self._ldap_datastore._database
        account = self._ldap_datastore._account_class({
            'uid': 'kgtestuser3',
            'givenName': 'Test',
            'sn': "User3",
            'cn': 'Test User3',
            'gidNumber': 500,
            'o': 'Example',
        })
        tldap.database.insert(account, database=database)

        group = self._ldap_datastore._group_class({
            'cn': 'Example',
            'gidNumber': 500,
        })
        tldap.database.insert(group, database=database)

        group = self._ldap_datastore._group_class({
            'cn': 'otherinst',
            'gidNumber': 501,
        })
        tldap.database.insert(group, database=database)

        group = self._ldap_datastore._group_class({
            'cn': 'TestProject1',
            'memberUid': ['kgtestuser3'],
            'gidNumber': 504,
        })
        tldap.database.insert(group, database=database)

        _DATASTORES.clear()
        _DATASTORES.append(self._ldap_datastore)

    def cleanup(self):
        tldap.transaction.rollback()
        tldap.transaction.leave_transaction_management()

        reset()
        _DATASTORES.clear()



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.conf.urls import include, url

from karaage.plugins.kgapplications.views import common, project

from .views.project import register as register_project


register_project()

urlpatterns = [
    url(r'^$',
        common.application_list, name='kg_application_list'),
    url(r'^applicants/(?P<applicant_id>\d+)/$',
        common.applicant_edit, name='kg_applicant_edit'),
    url(r'^(?P<application_id>\d+)/logs/$',
        common.application_logs, name='kg_application_logs'),
    url(r'^(?P<application_id>\d+)/add_comment/$',
        common.add_comment, name='kg_application_add_comment'),

    url(r'^(?P<application_id>\d+)/$',
        common.application_detail, name='kg_application_detail'),
    url(r'^(?P<application_id>\d+)/(?P<state>[-.\w]+)/$',
        common.application_detail, name='kg_application_detail'),
    url(r'^(?P<application_id>\d+)/(?P<state>[-.\w]+)/(?P<label>[-.\w]+)/$',
        common.application_detail, name='kg_application_detail'),

    url(r'^project/new/$',
        project.new_application, name='kg_application_new'),
    url(r'^project/invite/$',
        project.send_invitation, name='kg_application_invite'),
    url(r'^project/invite/(?P<project_id>\d+)/$',
        project.send_invitation, name='kg_application_invite'),

    # this must come last
    url(r'^(?P<token>[-.\w]+)/$',
        common.application_unauthenticated,
        name='kg_application_unauthenticated'),
    url(r'^(?P<token>[-.\w]+)/(?P<state>[-.\w]+)/$',
        common.application_unauthenticated,
        name='kg_application_unauthenticated'),
    url(r'^(?P<token>[-.\w]+)/(?P<state>[-.\w]+)/(?P<label>[-.\w]+)/$',
        common.application_unauthenticated,
        name='kg_application_unauthenticated'),
]

urlpatterns = [
    url(r'^applications/', include(urlpatterns)),
]

profile_urlpatterns = [
    url(r'^applications/$',
        common.profile_application_list,
        name='kg_profile_applications'),
]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" Base file used for all datastores. """


class DataStore(object):
    """ Base class used for all MC datastores. """

    def __init__(self, config):
        self.config = config

    ###########
    # ACCOUNT #
    ###########

    def save_account(self, account):
        """ Account was saved. """
        return

    def delete_account(self, account):
        """ Account was deleted. """
        return

    def set_account_password(self, account, raw_password):
        """ Account's password was changed. """
        return

    def set_account_username(self, account, old_username, new_username):
        """ Account's username was changed. """
        return

    def add_account_to_group(self, account, group):
        """ Add account to group. """
        return

    def remove_account_from_group(self, account, group):
        """ Remove account from group. """
        return

    def add_account_to_project(self, account, project):
        """ Add account to project. """
        return

    def remove_account_from_project(self, account, project):
        """ Remove account from project. """
        return

    def add_account_to_institute(self, account, institute):
        """ Add account to institute. """
        return

    def remove_account_from_institute(self, account, institute):
        """ Remove account from institute. """
        return

    def account_exists(self, username):
        """ Does the account exist? """
        return False

    def get_account_details(self, account):
        """ Get the account details """
        return {}

    #########
    # GROUP #
    #########

    def save_group(self, group):
        """ Group was saved. """
        return

    def delete_group(self, group):
        """ Group was deleted. """
        return

    def set_group_name(self, group, old_name, new_name):
        """ Group was renamed. """
        return

    def get_group_details(self, group):
        """ Get the group details. """
        return {}

    ###########
    # PROJECT #
    ###########

    def save_project(self, project):
        """ Project was saved. """
        return

    def delete_project(self, project):
        """ Project was deleted. """
        return

    def get_project_details(self, project):
        """ Get project's details. """
        return {}

    def set_project_pid(self, project, old_pid, new_pid):
        """ Project's pid was changed. """
        return

    #############
    # INSTITUTE #
    #############

    def save_institute(self, institute):
        """ Institute was saved. """
        return

    def delete_institute(self, institute):
        """ Institute was deleted. """
        return

    def get_institute_details(self, institute):
        """ Get institute's details. """
        return {}



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

# Django 1.7 style AppConfig, see
# http://docs.djangoproject.com/en/1.7/ref/applications/
from .celery import app as celery_app  # NOQA


default_app_config = 'karaage.apps.Karaage'

__author__ = """Brian May"""
__email__ = 'brian@linuxpenguins.xyz'
__version__ = '5.0.14'



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django.contrib import auth
from django.core.exceptions import ImproperlyConfigured
from django.shortcuts import render
from django.utils.deprecation import MiddlewareMixin

import karaage.common.saml as saml
from karaage.people.models import Person


class SamlUserMiddleware(MiddlewareMixin):
    """
    Middleware for utilizing Web-server-provided authentication.

    If request.user is not authenticated, then this middleware attempts to
    authenticate the username passed in the ``HTTP_PERSISTENT_ID`` request
    header.  If authentication is successful, the user is automatically logged
    in to persist the user in the session.

    The header used is configurable and defaults to ``HTTP_PERSISTENT_ID``.
    Subclass this class and change the ``header`` attribute if you need to use
    a different header.
    """

    def process_request(self, request):
        # AuthenticationMiddleware is required so that request.user exists.
        if not hasattr(request, 'user'):
            raise ImproperlyConfigured(
                "The Django SAML user auth middleware requires the"
                " authentication middleware to be installed.  Edit your"
                " MIDDLEWARE_CLASSES setting to insert"
                " 'django.contrib.auth.middleware.AuthenticationMiddleware'"
                " before the SamlUserMiddleware class.")

        # If the user is already authenticated and that user is the user we are
        # getting passed in the headers, then the correct user is already
        # persisted in the session and we don't need to continue.
        if request.user.is_authenticated:
            return

        # Is this a shib session?
        if not saml.is_saml_session(request):
            return

        # Can we get the shib attributes we need?
        attrs, error = saml.parse_attributes(request)
        if error:
            return render(template_name='saml_error.html',
                          context={'shib_attrs': attrs},
                          request=request)

        # What is our persistent_id?
        saml_id = attrs['persistent_id']
        assert saml_id

        # We are seeing this user for the first time in this session, attempt
        # to authenticate the user.
        try:
            person = Person.objects.get(saml_id=saml_id)
        except Person.DoesNotExist:
            return

        # User is valid.  Set request.user and persist user in the session
        # by logging the user in.
        request.user = person
        # We must set the model backend here manually as we skip
        # the call to auth.authenticate().
        request.user.backend = 'django.contrib.auth.backends.ModelBackend'
        auth.login(request, person)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from __future__ import absolute_import

import json
import logging
from math import ceil

from alogger import get_parser

from karaage.machines.models import Account, Machine
from karaage.projects.models import Project

from .models import CPUJob, Queue


logger = logging.getLogger(__name__)

"""
Parse log files using alogger.
"""


class AloggerParser(object):

    def line_to_dict(self, line):
        return json.loads(line)


def parse_logs(log_list, date, machine_name, log_type):
    """
    Parse log file lines in log_type format.
    """
    output = []
    count = fail = skip = updated = 0

    # Check things are setup correctly
    try:
        machine = Machine.objects.get(name=machine_name)
    except Machine.DoesNotExist:
        return "ERROR: Couldn't find machine named: %s" % machine_name

    if log_type == "alogger":
        parser = AloggerParser()
    else:
        parser = get_parser(log_type)

    # Process each line
    for line_no, line in enumerate(log_list):
        try:
            data = parser.line_to_dict(line)
        except ValueError:
            output.append("%d: Error reading line" % line_no)
            continue

        # if parser returns None, nothing to do, continue to next line
        if data is None:
            skip += 1
            continue

        # check for required fields
        required_fields = [
            'user', 'project', 'jobid', 'jobname',
            'cpu_usage', 'cores',
            'act_wall_time', 'est_wall_time',
            'mem', 'vmem', 'list_pmem', 'list_mem', 'list_pvmem',
            'ctime', 'qtime', 'etime', 'start',
        ]

        for field in required_fields:
            if field not in data:
                output.append(
                    "line %d: %s field not given." % (line_no, field))
                fail = fail + 1
                continue

        # Process user --> account
        try:
            account = Account.objects.get(
                username=data['user'],
                date_deleted__isnull=True)

        except Account.DoesNotExist:
            # Couldn't find user account - Assign to user None
            output.append(
                "line %d: Couldn't find user account for username=%s."
                % (line_no, data['user']))
            fail += 1
            continue

        except Account.MultipleObjectsReturned:
            output.append(
                "line %d: Username %s has multiple active accounts."
                % (line_no, data['user']))
            fail += 1
            continue

        # Process project
        if data['project'] is None:
            output.append(
                "line %d: Project was not supplied." % (line_no))
            fail += 1
            continue

        try:
            project = Project.objects.get(pid=data['project'])

        except Project.DoesNotExist:
            output.append(
                "line %d: Couldn't find specified project %s"
                % (line_no, data['project']))
            fail += 1
            continue

        # memory calculations
        if machine.mem_per_core:
            avail_mem_per_core = machine.mem_per_core * 1024
            avail_mem_for_job = avail_mem_per_core * data['cores']

            if data['list_pmem'] * data['cores'] > data['list_mem']:
                memory_used_per_core = data['list_pmem']
                memory_used_for_job = data['list_pmem'] * data['cores']
            else:
                memory_used_per_core = data['list_mem'] / data['cores']
                memory_used_for_job = data['list_mem']

            if memory_used_for_job > avail_mem_for_job:
                data['cpu_usage'] = ceil(
                    memory_used_per_core / avail_mem_per_core
                    * data['act_wall_time']
                    * data['cores'])

        # apply scaling factor to cpu_usage
        data['cpu_usage'] = data['cpu_usage'] * machine.scaling_factor

        # Everything is good so add entry
        queue, created = Queue.objects.get_or_create(name=data['queue'])

        try:
            cpujob, created = CPUJob.objects.get_or_create(jobid=data['jobid'])
            cpujob.account = account
            cpujob.username = data['user']
            cpujob.project = project
            cpujob.machine = machine
            cpujob.date = date
            cpujob.queue = queue
            cpujob.cpu_usage = data['cpu_usage']
            cpujob.est_wall_time = data['est_wall_time']
            cpujob.act_wall_time = data['act_wall_time']
            cpujob.mem = data['mem']
            cpujob.vmem = data['vmem']
            cpujob.ctime = data['ctime']
            cpujob.qtime = data['qtime']
            cpujob.etime = data['etime']
            cpujob.start = data['start']
            cpujob.cores = data['cores']
            cpujob.exit_status = data['exit_status']
            cpujob.jobname = data['jobname']
            cpujob.list_mem = data['list_mem']
            cpujob.list_vmem = data['list_vmem']
            cpujob.list_pmem = data['list_pmem']
            cpujob.list_pvmem = data['list_pvmem']
            cpujob.save()

        except Exception as e:
            output.append(
                "line %d: Failed to insert a line  - %s" % (line_no, e))
            fail += 1
            continue

        if created:
            count += 1
        else:
            updated += 1

    summary = (
        'Inserted : %i\nUpdated  : %i\nFailed   : %i\nSkiped   : %i'
        % (count, updated, fail, skip)
    )

    logger.debug('Inserted : %i' % count)
    logger.debug('Updated  : %i' % updated)
    logger.debug('Failed   : %i' % fail)
    logger.debug('Skiped   : %i' % skip)

    return summary, output



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

"""
Management utility to import user from a CSV file.

CSV format should be:

username,password,short_name,full_name,email,institute,project
sam,secret,Joe,Joe Bloggs,joe@example.com,Test,TestProject2
bob,secret2,Bob,Bob Smith,bob@example.com,Example University,pEx0032


"""

import re
import sys
from csv import DictReader

import django.db.transaction
import tldap.transaction
from django.core import exceptions
from django.core.management.base import BaseCommand
from django.core.validators import validate_email

from karaage.institutes.models import Institute
from karaage.people.models import Person
from karaage.projects.models import Project
from karaage.projects.utils import add_user_to_project


RE_VALID_USERNAME = re.compile(r'[\w.@+-]+$')


class Command(BaseCommand):
    help = """Import users from a CSV file with the following format.
username,password,short_name,full_name,email,institute,project"""
    args = "csvfile"

    @django.db.transaction.atomic
    @tldap.transaction.commit_on_success
    def handle(self, csvfile, **options):
        verbosity = int(options.get('verbosity', 1))

        try:
            data = DictReader(open(csvfile))
        except FileNotFoundError:
            sys.stderr.write("ERROR: Failed to read CSV file.\n")
            sys.exit(1)

        success = 0
        fail_count = 0
        skip = 0
        for user in data:
            fail = False

            if verbosity >= 1:
                print("Attempting to import user '%s'" % user['username'])

            if 'username' not in user:
                sys.stderr.write("Error: Failed to find username column.\n")
                fail = True
            if 'password' not in user:
                sys.stderr.write("Error: Failed to find password column.\n")
                fail = True
            if 'short_name' not in user:
                sys.stderr.write("Error: Failed to find short_name column.\n")
                fail = True
            if 'full_name' not in user:
                sys.stderr.write("Error: Failed to find full_name column.\n")
                fail = True
            if 'email' not in user:
                sys.stderr.write("Error: Failed to find email column.\n")
                fail = True
            if 'institute' not in user:
                sys.stderr.write("Error: Failed to find institute column.\n")
                fail = True
            if 'project' not in user:
                sys.stderr.write("Error: Failed to find project column.\n")
                fail = True

            if not RE_VALID_USERNAME.match(user['username']):
                sys.stderr.write(
                    "Error: Username is invalid. "
                    "Use only letters, digits and underscores.\n")
                fail = True

            try:
                validate_email(user['email'])
            except exceptions.ValidationError:
                sys.stderr.write(
                    "Error: E-mail address '%s' is invalid.\n" % user['email'])
                fail = True

            if fail:
                sys.stderr.write(
                    "Skipping row for username '%s' due to errors\n"
                    % user['username'])
                fail_count += 1
                continue

            try:
                Person.objects.get(username=user['username'])
                sys.stderr.write(
                    "Error: Username '%s' exists. Skipping\n"
                    % user['username'])
                skip += 1
                continue
            except Person.DoesNotExist:
                pass

            try:
                institute = Institute.objects.get(name=user['institute'])
                user['institute'] = institute
            except Institute.DoesNotExist:
                sys.stderr.write(
                    "Error: Institute '%s' does not exist. Skipping\n"
                    % user['institute'])
                fail_count += 1
                continue

            project = None
            if user['project']:
                try:
                    project = Project.objects.get(pid=user['project'])
                except Project.DoesNotExist:
                    sys.stderr.write(
                        "Error: Project '%s' does not exist. Skipping\n"
                        % user['project'])
                    fail_count += 1
                    continue

            user['password1'] = user['password']
            person = Person.objects.create_user(**user)
            print("Successfully added user '%s'" % person)
            if project:
                add_user_to_project(person, project)

            success += 1

        print('')
        print('Added:   %s' % success)
        print('Skipped: %s' % skip)
        print('Failed:  %s' % fail_count)

        sys.exit(0)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django.views.static
import django_xmlrpc.views
from django.conf import settings
from django.conf.urls import include, url

from karaage.common import get_urls
from karaage.common.views import common


# Profile URLS

profile_urlpatterns = [
]


def _load_profile_urls():
    import importlib
    global profile_urlpatterns
    modules = [
        'karaage.institutes.urls',
        'karaage.projects.urls',
        'karaage.people.urls',
        'karaage.machines.urls',
    ]
    for module_name in modules:
        module = importlib.import_module(module_name)
        profile_urlpatterns += module.profile_urlpatterns


_load_profile_urls()

for urls in get_urls("profile_urlpatterns"):
    profile_urlpatterns += urls
    del urls


# Standard URLS

urlpatterns = [
    url(r'^xmlrpc/$', django_xmlrpc.views.handle_xmlrpc),
    url(r'^captcha/', include('captcha.urls')),
    url(r'^lookup/', include('ajax_select.urls')),

    url(r'^emails/', include('karaage.emails.urls')),
    url(r'^institutes/', include('karaage.institutes.urls')),
    url(r'^projects/', include('karaage.projects.urls')),
    url(r'^persons/', include('karaage.people.urls.persons')),
    url(r'^groups/', include('karaage.people.urls.groups')),
    url(r'^accounts/', include('karaage.machines.urls.accounts')),
    url(r'^machines/', include('karaage.machines.urls.machines')),
    url(r'^profile/', include(profile_urlpatterns)),

    url(r'^$', common.index, name='index'),
    url(r'^search/$', common.search,
        name='kg_site_search'),
    url(r'^misc/$', common.misc, name='kg_misc'),
    url(r'^logs/$', common.log_list,
        name='kg_log_list'),
    url(r'^aup/$', common.aup, name="kg_aup"),
]

for urls in get_urls("urlpatterns"):
    urlpatterns += urls
    del urls

if settings.DEBUG:
    urlpatterns += [
        url(r'^kgfiles/(?P<path>.*)$', django.views.static.serve,
            {'document_root': settings.FILES_DIR})
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import sys

import django.db.transaction
import tldap.transaction
from django.conf import settings
from django.core.management.base import BaseCommand

from karaage.people.models import Person


# Generate a human readable 'random' password
# password  will be generated in the form 'word'+digits+'word'
# eg.,nice137pass
# parameters: number of 'characters' , number of 'digits'
# Pradeep Kishore Gowda <pradeep at btbytes.com >
# License : GPL
# Date : 2005.April.15
# Revision 1.2
# ChangeLog:
# 1.1 - fixed typos
# 1.2 - renamed functions _apart & _npart to a_part
#       & n_part as zope does not allow functions to
# start with _
def nicepass(alpha=8, numeric=4):
    """
    returns a human-readble password (say rol86din instead of
    a difficult to remember K8Yn9muL )
    """
    import string
    import random
    vowels = ['a', 'e', 'i', 'o', 'u']
    consonants = [a for a in string.ascii_lowercase if a not in vowels]
    digits = string.digits

    # utility functions
    def a_part(slen):
        ret = ''
        for i in range(slen):
            if i % 2 == 0:
                randid = random.randint(0, 20)  # number of consonants
                ret += consonants[randid]
            else:
                randid = random.randint(0, 4)  # number of vowels
                ret += vowels[randid]
        return ret

    def n_part(slen):
        ret = ''
        for i in range(slen):
            randid = random.randint(0, 9)  # number of digits
            ret += digits[randid]
        return ret

    ####
    fpl = alpha / 2
    if alpha % 2:
        fpl = int(alpha / 2) + 1
    lpl = alpha - fpl

    start = a_part(fpl)
    mid = n_part(numeric)
    end = a_part(lpl)

    return "%s%s%s" % (start, mid, end)


class Command(BaseCommand):
    help = "Unlock all training accounts and reset password."

    def add_arguments(self, parser):
        parser.add_argument(
            '--password', dest="password",
            action='store_true', default=False,
            help="Read password to use on stdin."),
        parser.add_argument(
            '--number', dest="number",
            type=int, help="Number of accounts to unlock"),

    @django.db.transaction.atomic
    @tldap.transaction.commit_on_success
    def handle(self, *args, **options):
        verbose = int(options.get('verbosity'))
        training_prefix = getattr(settings, 'TRAINING_ACCOUNT_PREFIX', 'train')

        # If training accounts are system users, they will be found by
        # Person.objects.all() but not Person.active.all()
        query = Person.objects.all()
        query = query.filter(username__iregex=training_prefix)
        query = query.order_by('username')

        if options['number'] is not None:
            query = query[:options['number']]

        password = None
        if options['password']:
            password = sys.stdin.readline().strip()
        else:
            password = nicepass(8, 4)
            print("New password: %s" % password)

        for person in query.all():
            try:
                if password is not None:
                    if verbose > 1:
                        print("%s: Setting password" % person.username)
                    person.set_password(password)
                    # person.unlock() will call person.save()

                if verbose > 1:
                    print("%s: Unlocking" % person.username)
                person.unlock()

            except Exception as e:
                print("%s: Failed to unlock: %s" % (person.username, e))



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2017-12-15 18:31
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('karaage', '0004_auto_20160429_0927'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='institutequota',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='institutequota',
            name='institute',
        ),
        migrations.RemoveField(
            model_name='institutequota',
            name='machine_category',
        ),
        migrations.AlterUniqueTogether(
            name='projectquota',
            unique_together=set([]),
        ),
        migrations.RemoveField(
            model_name='projectquota',
            name='machine_category',
        ),
        migrations.RemoveField(
            model_name='projectquota',
            name='project',
        ),
        migrations.RemoveField(
            model_name='account',
            name='machine_category',
        ),
        migrations.RemoveField(
            model_name='machine',
            name='category',
        ),
        migrations.DeleteModel(
            name='MachineCategory',
        ),
        migrations.DeleteModel(
            name='ProjectQuota',
        ),
        migrations.DeleteModel(
            name='InstituteQuota',
        ),
    ]



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

from django_xmlrpc.decorators import xmlrpc_func

from karaage.machines.models import Account


@xmlrpc_func(returns='int', args=['string', 'string'])
def get_disk_quota(username, machine_name=None):
    """
    Returns disk quota for username in KB
    """

    try:
        ua = Account.objects.get(
            username=username,
            date_deleted__isnull=True)
    except Account.DoesNotExist:
        return 'Account not found'

    result = ua.get_disk_quota()
    if result is None:
        return False

    return result * 1048576



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

""" LDAP datastore. """

import importlib
import logging

import tldap.backend
from tldap import Q
from tldap.database import (
    Database,
    LdapObject,
    LdapObjectClass,
    changeset,
    delete,
    get_one,
    preload,
    rename,
    save,
)
from tldap.exceptions import ObjectDoesNotExist

import karaage.common.trace as trace
from karaage.datastores import base
from karaage.datastores.ldap_schemas import OpenldapAccount, OpenldapGroup
from karaage.machines.models import Account


logger = logging.getLogger(__name__)


def _str_or_none(string):
    """ Return a string of None if string is empty. """
    if string is None or string == "":
        return None
    return string


def _lookup(cls: str) -> LdapObjectClass:
    """ Lookup module.class. """
    if isinstance(cls, str):
        module_name, _, name = cls.rpartition(".")
        module = importlib.import_module(module_name)
        try:
            cls = getattr(module, name)
        except AttributeError:
            raise AttributeError("%s reference cannot be found" % cls)
    return cls


class DataStore(base.DataStore):
    """ LDAP Account and group datastore. """

    def __init__(self, config: dict) -> None:
        super(DataStore, self).__init__(config)
        using = config['LDAP']

        connection = tldap.backend.connections[using]
        self._database = Database(connection, config)

        self._account_class = _lookup(config['ACCOUNT'])
        self._group_class = _lookup(config['GROUP'])
        self._primary_group = config.get('PRIMARY_GROUP', 'institute')
        self._default_primary_group = config.get('DEFAULT_PRIMARY_GROUP', 'dummy')
        self._settings = config

    def _get_account(self, uid: str) -> LdapObject:
        return get_one(
            table=self._account_class,
            query=Q(uid=uid),
            database=self._database,
        )

    def _get_group(self, cn: str) -> LdapObject:
        return get_one(
            table=self._group_class,
            query=Q(cn=cn),
            database=self._database,
        )

    def save_account(self, account: Account) -> None:
        """ Account was saved. """
        person = account.person
        if self._primary_group == 'institute':
            lgroup = self._get_group(person.institute.group.name)
        elif self._primary_group == 'default_project':
            if account.default_project is None:
                lgroup = self._get_group(self._default_primary_group)
            else:
                lgroup = self._get_group(account.default_project.group.name)
        else:
            raise RuntimeError("Unknown value of PRIMARY_GROUP.")

        if account.default_project is None:
            default_project = "none"
        else:
            default_project = account.default_project.pid

        try:
            luser = self._get_account(account.username)
            changes = changeset(luser, {})
            new_user = False
        except ObjectDoesNotExist:
            new_user = True
            luser = self._account_class()
            changes = changeset(luser, {
                'uid': account.username
            })

        changes = changes.merge({
            'gidNumber': lgroup['gidNumber'],
            'givenName': person.first_name,
            'sn': person.last_name,
            'telephoneNumber': _str_or_none(person.telephone),
            'mail': _str_or_none(person.email),
            'title': _str_or_none(person.title),
            'o': person.institute.name,
            'cn': person.full_name,
            'default_project': default_project,
            'loginShell': account.shell,
            'locked': account.is_locked()
        })
        save(changes, database=self._database)

        if new_user:
            # add all groups
            for group in account.person.groups.all():
                self.add_account_to_group(account, group)

    def delete_account(self, account):
        """ Account was deleted. """
        try:
            luser = self._get_account(account.username)
            groups = luser['groups'].load(database=self._database)
            for group in groups:
                changes = changeset(group, {})
                changes = group.remove_member(changes, luser)
                save(changes, database=self._database)

            delete(luser, database=self._database)
        except ObjectDoesNotExist:
            # it doesn't matter if it doesn't exist
            pass

    def set_account_password(self, account, raw_password):
        """ Account's password was changed. """
        luser = self._get_account(account.username)
        changes = changeset(luser, {
            'password': raw_password,
        })
        save(changes, database=self._database)

    def set_account_username(self, account, old_username, new_username):
        """ Account's username was changed. """
        luser = self._get_account(old_username)
        rename(luser, database=self._database, uid=new_username)

    def add_account_to_group(self, account, group):
        """ Add account to group. """
        lgroup: OpenldapGroup = self._get_group(group.name)
        person: OpenldapAccount = self._get_account(account.username)

        changes = changeset(lgroup, {})
        changes = lgroup.add_member(changes, person)
        save(changes, database=self._database)

    def remove_account_from_group(self, account, group):
        """ Remove account from group. """
        lgroup: OpenldapGroup = self._get_group(group.name)
        person: OpenldapAccount = self._get_account(account.username)

        changes = changeset(lgroup, {})
        changes = lgroup.remove_member(changes, person)
        save(changes, database=self._database)

    def get_account_details(self, account):
        """ Get the account details. """
        result = {}
        try:
            luser = self._get_account(account.username)
            luser = preload(luser, database=self._database)
        except ObjectDoesNotExist:
            return result

        for i, j in luser.items():
            if i != 'userPassword' and j is not None:
                result[i] = j

        return result

    def account_exists(self, username):
        """ Does the account exist? """
        try:
            self._get_account(username)
            return True
        except ObjectDoesNotExist:
            return False

    def save_group(self, group):
        """ Group was saved. """
        # If group already exists, take over existing group rather then error.
        try:
            lgroup = self._get_group(group.name)
            changes = changeset(lgroup, {})
        except ObjectDoesNotExist:
            lgroup = self._group_class()
            changes = changeset(lgroup, {
                'cn': group.name,
            })

        changes = changes.merge({
            'description': group.description
        })
        save(changes, database=self._database)

    def delete_group(self, group):
        """ Group was deleted. """
        try:
            lgroup = self._get_group(group.name)
            delete(lgroup, database=self._database)
        except ObjectDoesNotExist:
            # it doesn't matter if it doesn't exist
            pass

    def set_group_name(self, group, old_name, new_name):
        """ Group was renamed. """
        lgroup = self._get_group(old_name)
        rename(lgroup, database=self._database, cn=new_name)

    def get_group_details(self, group):
        """ Get the group details. """
        result = {}
        try:
            lgroup = self._get_group(group.name)
            lgroup = preload(lgroup, database=self._database)
        except ObjectDoesNotExist:
            return result

        for i, j in lgroup.items():
            if j is not None:
                result[i] = j

        return result


trace.attach(trace.trace(logger), DataStore)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import django
import pytest
from django.conf import settings
from django.core import mail
from django.core.management import call_command
from django.test import TestCase
from django.urls import reverse

from karaage.people.models import Group
from karaage.plugins.kgapplications.models import Application

from ...models import Software, SoftwareLicense
from ..models import SoftwareApplication


def set_admin():
    settings.ADMIN_IGNORED = False


def set_no_admin():
    settings.ADMIN_IGNORED = True


@pytest.mark.django_db
class SoftwareApplicationTestCase(TestCase):

    def setUp(self):
        call_command('loaddata', 'test_karaage', **{'verbosity': 0})

    def tearDown(self):
        set_admin()

    def test_register_software(self):
        if django.VERSION >= (1, 9):
            url_prefix = ""
        else:
            url_prefix = 'http://testserver'

        group = Group.objects.create(name="windows")
        software = Software.objects.create(
            name="windows",
            restricted=True,
            group=group,
        )
        SoftwareLicense.objects.create(
            software=software,
            version="3.11",
            text="You give your soal to the author "
            "if you wish to access this software.",
        )

        set_no_admin()

        # APPLICANT LOGS IN
        logged_in = self.client.login(
            username='kgtestuser1', password='aq12ws')
        self.assertEqual(logged_in, True)
        self.assertEqual(len(mail.outbox), 0)

        response = self.client.get(
            reverse('kg_software_detail', args=[software.pk]))
        self.assertEqual(response.status_code, 200)

        # OPEN APPLICATION
        form_data = {
        }

        response = self.client.post(
            reverse('kg_software_detail', args=[software.pk]),
            form_data, follow=True)
        self.assertEqual(response.status_code, 200)
        application = Application.objects.get()
        self.assertEqual(
            response.redirect_chain[0][0],
            url_prefix
            + reverse('kg_application_detail', args=[application.pk, 'O']))
        self.assertEqual(len(mail.outbox), 1)
        self.assertEqual(
            mail.outbox[0].subject,
            'TestOrg invitation: access software windows')
        self.assertEqual(mail.outbox[0].from_email, settings.ACCOUNTS_EMAIL)
        self.assertEqual(mail.outbox[0].to[0], 'leader@example.com')

        # SUBMIT APPLICATION
        form_data = {
            'submit': True,
        }

        response = self.client.post(
            reverse('kg_application_detail', args=[application.pk, 'O']),
            form_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.redirect_chain[0][0],
            url_prefix
            + reverse('kg_application_detail', args=[application.pk, 'K']))

        self.assertEqual(len(mail.outbox), 2)
        self.assertEqual(
            mail.outbox[1].subject, 'TestOrg request: access software windows')
        self.assertEqual(mail.outbox[1].from_email, settings.ACCOUNTS_EMAIL)
        self.assertEqual(mail.outbox[1].to[0], 'sam@vpac.org')

        # ADMIN LOGS IN TO APPROVE
        set_admin()
        logged_in = self.client.login(username='kgsuper', password='aq12ws')
        self.assertEqual(logged_in, True)

        # ADMIN GET DETAILS
        response = self.client.get(
            reverse('kg_application_detail', args=[application.pk, 'K']))
        self.assertEqual(response.status_code, 200)

        # ADMIN GET DECLINE PAGE
        response = self.client.get(
            reverse('kg_application_detail',
                    args=[application.pk, 'K', 'cancel']))
        self.assertEqual(response.status_code, 200)

        # ADMIN GET APPROVE PAGE
        response = self.client.get(
            reverse('kg_application_detail',
                    args=[application.pk, 'K', 'approve']))
        self.assertEqual(response.status_code, 200)

        # ADMIN APPROVE
        form_data = {
            'make_leader': False,
            'additional_req': 'Woof',
            'needs_account': False,
            'approve': True,
        }
        response = self.client.post(
            reverse('kg_application_detail',
                    args=[application.pk, 'K', 'approve']),
            form_data, follow=True)
        self.assertEqual(response.status_code, 200)
        self.assertEqual(
            response.redirect_chain[0][0],
            url_prefix
            + reverse('kg_application_detail', args=[application.pk, 'C']))
        application = Application.objects.get(pk=application.id)
        self.assertEqual(application.state, SoftwareApplication.COMPLETED)
        self.assertEqual(len(mail.outbox), 3)
        self.assertEqual(
            mail.outbox[2].subject,
            'TestOrg approved: access software windows')
        self.assertEqual(mail.outbox[2].from_email, settings.ACCOUNTS_EMAIL)
        self.assertEqual(mail.outbox[2].to[0], 'leader@example.com')
        self.client.logout()
        set_no_admin()

        # test group
        groups = Group.objects.filter(
            name="windows", members__username="kgtestuser1")
        self.assertEqual(len(groups), 1)



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import ajax_select.fields
import six
from django import forms
from django.conf import settings
from django.contrib.admin.widgets import AdminDateWidget
from django.contrib.auth.forms import SetPasswordForm as BaseSetPasswordForm

from karaage.common import get_current_person
from karaage.common.constants import COUNTRIES
from karaage.common.forms import clean_email, validate_password
from karaage.institutes.models import Institute
from karaage.people.models import Group, Person
from karaage.people.utils import (
    UsernameException,
    validate_username_for_new_person,
)
from karaage.projects.models import Project
from karaage.projects.utils import add_user_to_project


class PersonForm(forms.ModelForm):
    #    title = forms.ChoiceField(choices=TITLES, required=False)
    #    position = forms.CharField(required=False)
    #    email = forms.EmailField()
    #    department = forms.CharField(required=False)
    #    supervisor = forms.CharField(required=False)
    telephone = forms.RegexField(
        r"^[0-9a-zA-Z\.( )+-]+$", required=True,
        label=six.u("Office Telephone"),
        help_text=six.u(
            "Used for emergency contact and password reset service."),
        error_messages={
            'invalid': 'Telephone number may only contain digits, letter, '
                       'hyphens, spaces, braces,  and the plus sign.'})
    mobile = forms.RegexField(
        "^[0-9a-zA-Z( )+-]+$",
        required=False,
        error_messages={
            'invalid': 'Telephone number may only contain digits, letter, '
                       'hyphens, spaces, braces,  and the plus sign.'})
    fax = forms.CharField(required=False)
    address = forms.CharField(
        label=six.u("Mailing Address"),
        required=False,
        widget=forms.Textarea())
    country = forms.ChoiceField(
        choices=COUNTRIES, initial='AU', required=False)

    def __init__(self, *args, **kwargs):
        super(PersonForm, self).__init__(*args, **kwargs)
        self.fields['short_name'].help_text = \
            "This is typically the person's given name. "\
            "For example enter 'Fred' here."
        self.fields['full_name'].help_text = \
            "This is typically the person's full name. " \
            "For example enter 'Fred Smith' here."

    def clean(self):
        data = super(PersonForm, self).clean()

        for key in [
                'short_name', 'full_name', 'email', 'position',
                'supervisor', 'department', 'telephone', 'mobile', 'fax',
                'address', ]:
            if key in data and data[key]:
                data[key] = data[key].strip()

        return data

    class Meta:
        model = Person
        fields = [
            'short_name', 'full_name', 'email', 'title', 'position',
            'supervisor', 'department', 'telephone', 'mobile', 'fax',
            'address', 'country'
        ]

    def clean_email(self):
        email = self.cleaned_data['email']
        users = Person.objects.filter(email__exact=email)
        if self.instance:
            users = users.exclude(pk=self.instance.pk)
        if users.count() > 0:
            raise forms.ValidationError(
                six.u(
                    'An account with this email already exists. '
                    'Please email %s')
                % settings.ACCOUNTS_EMAIL)
        clean_email(email)
        return email


class AdminPersonForm(PersonForm):
    institute = forms.ModelChoiceField(queryset=None)
    comment = forms.CharField(widget=forms.Textarea(), required=False)
    expires = forms.DateField(widget=AdminDateWidget, required=False)
    is_admin = forms.BooleanField(
        help_text="Designates whether the user can log into this admin site.",
        required=False)
    is_systemuser = forms.BooleanField(
        help_text="Designates that this user is a system process, "
                  "not a person.",
        required=False)

    def __init__(self, *args, **kwargs):
        super(AdminPersonForm, self).__init__(*args, **kwargs)
        self.fields['institute'].queryset = Institute.active.all()

    class Meta:
        model = Person
        fields = [
            'short_name', 'full_name', 'email', 'title', 'position',
            'supervisor', 'department', 'institute', 'telephone', 'mobile',
            'fax', 'address', 'country', 'expires', 'comment',
            'is_systemuser', 'is_admin', ]


class AddPersonForm(AdminPersonForm):
    project = forms.ModelChoiceField(
        queryset=None,
        label=six.u("Default Project"), required=False)
    needs_account = forms.BooleanField(
        required=False, label=six.u("Do you require a cluster account"),
        help_text=six.u("eg. Will you be working on the project yourself"))
    username = forms.CharField(
        label=six.u("Requested username"),
        max_length=settings.USERNAME_MAX_LENGTH,
        help_text=(settings.USERNAME_VALIDATION_ERROR_MSG
                   + " and has a max length of %s."
                   % settings.USERNAME_MAX_LENGTH))
    password1 = forms.CharField(
        widget=forms.PasswordInput(render_value=False),
        label=six.u('Password'))
    password2 = forms.CharField(
        widget=forms.PasswordInput(render_value=False),
        label=six.u('Password (again)'))

    def __init__(self, *args, **kwargs):
        super(AddPersonForm, self).__init__(*args, **kwargs)
        self.fields['project'].queryset = Project.objects.all()

    def clean_username(self):
        username = self.cleaned_data['username']
        try:
            validate_username_for_new_person(username)
        except UsernameException as e:
            raise forms.ValidationError(e.args[0])
        return username

    def clean_password2(self):
        username = self.cleaned_data.get('username')
        password1 = self.cleaned_data.get('password1')
        password2 = self.cleaned_data.get('password2')
        return validate_password(username, password1, password2)

    def save(self, commit=True):
        assert commit is True

        data = self.cleaned_data

        person = super(AddPersonForm, self).save(commit=False)
        person.username = data['username']
        person.is_admin = data['is_admin']
        person.is_active = True
        person.approved_by = get_current_person()
        person.set_password(data['password2'])
        person.save()

        if data['needs_account'] and data['project']:
            add_user_to_project(person, data['project'])

        return person


class AdminPasswordChangeForm(forms.Form):
    new1 = forms.CharField(
        widget=forms.PasswordInput(),
        label=six.u('New Password'))
    new2 = forms.CharField(
        widget=forms.PasswordInput(),
        label=six.u('New Password (again)'))

    def __init__(self, person, *args, **kwargs):
        self.person = person
        super(AdminPasswordChangeForm, self).__init__(*args, **kwargs)

    def clean_new2(self):
        username = self.person.username
        password1 = self.cleaned_data.get('new1')
        password2 = self.cleaned_data.get('new2')
        return validate_password(username, password1, password2)

    def save(self):
        data = self.cleaned_data
        person = self.person
        person.set_password(data['new1'])
        person.save()


class PasswordChangeForm(AdminPasswordChangeForm):
    old = forms.CharField(widget=forms.PasswordInput(), label='Old password')

    def clean_new2(self):
        username = self.person.username
        password1 = self.cleaned_data.get('new1')
        password2 = self.cleaned_data.get('new2')
        old_password = self.cleaned_data.get('old', None)
        return validate_password(username, password1, password2, old_password)

    def clean_old(self):
        person = Person.objects.authenticate(
            username=self.person.username,
            password=self.cleaned_data['old'])
        if person is None:
            raise forms.ValidationError(
                six.u('Your old password was incorrect'))

        return self.cleaned_data['old']


class SetPasswordForm(BaseSetPasswordForm):

    def clean_new_password1(self):
        password1 = self.cleaned_data.get('new_password1')
        return validate_password(self.user.username, password1)


class AdminGroupForm(forms.Form):
    name = forms.RegexField(
        "^%s$" % settings.GROUP_VALIDATION_RE,
        required=True,
        error_messages={'invalid': settings.GROUP_VALIDATION_ERROR_MSG})
    description = forms.CharField()

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.pop('instance', None)
        super(AdminGroupForm, self).__init__(*args, **kwargs)
        if self.instance is not None:
            self.initial = self.instance.__dict__

    def clean_name(self):
        name = self.cleaned_data["name"]
        groups = Group.objects.filter(name=name)
        if self.instance is not None:
            groups = groups.exclude(pk=self.instance.pk)
        if groups.count() > 0:
            raise forms.ValidationError("That group name already exists.")
        return name

    def save(self, group=None):
        data = self.cleaned_data

        if self.instance is None:
            group = Group()
        else:
            group = self.instance

        group.name = data['name']
        group.description = data['description']
        group.save()

        return group


class AddGroupMemberForm(forms.Form):

    """ Add a user to a group form """
    person = ajax_select.fields.AutoCompleteSelectField(
        'person', required=True, label="Add person")

    def __init__(self, *args, **kwargs):
        self.instance = kwargs.pop('instance', None)
        super(AddGroupMemberForm, self).__init__(*args, **kwargs)

    def save(self, commit=True):
        person = self.cleaned_data['person']
        self.instance.add_person(person)
        return self.instance



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import os
import shutil

from django.core.management.base import BaseCommand

from ...dirs import GRAPH_ROOT


class Command(BaseCommand):
    help = "Deletes all usage graphs"

    def handle(self, **options):
        verbose = int(options.get('verbosity'))

        if verbose > 1:
            print("Cleaning all graphs")

        for dirpath, dirnames, filenames in os.walk(GRAPH_ROOT):
            for name in dirnames:
                path = os.path.join(dirpath, name)
                if verbose > 1:
                    print("Cleaning %s" % path)
                shutil.rmtree(path)
            break



# Copyright 2010-2017, The University of Melbourne
# Copyright 2010-2017, Brian May
#
# This file is part of Karaage.
#
# Karaage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Karaage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Karaage  If not, see <http://www.gnu.org/licenses/>.

import tldap.ldap_passwd
from django.contrib.auth.backends import ModelBackend

from karaage.people.models import Person


class LDAPBackend(ModelBackend):

    def authenticate(self, username=None, password=None, **kwargs):
        try:
            person = Person.objects.get(username__exact=username)
        except Person.DoesNotExist:
            return None

        if person.legacy_ldap_password is None:
            return None

        if not tldap.ldap_passwd.check_password(
                password, person.legacy_ldap_password):
            return None

        # Success.
        person.set_password(password)
        person.save()

        return person
